"""
The `brad` module serves as the main interface for user interactions, whether through a graphical user interface (GUI), command line, or programmatically.

The `Agent` class creates a single chatbot instance that can be queried in various ways. 

The `AgentFactory` class is a session factory method, used to fetch and maintain Agent sessions from outside the module.  

The `brad.chat` method allows users to initiate a command line chat session without needing to create an `Agent` instance.

Main Methods
-------------

Main Methods:
    
1. `Agent.chat`
    This method creates a chat session where a user and `Agent` can have a conversation with back-and-forth inputs.
2. `Agent.invoke`
    This method responds to an individual user query with a single tool.
3. `AgentFactory.get_agent`
    Method to instantiate a new 

.. _state-schema-section:

State Schema
------------

The `Agent` state is managed within a dictionary called `Agent.state`. This dictionary tracks the agents inputs, outputs, memory, configurations, and more.
To pass this information between the `Agent` and each tool the `state` dictionary is passed as the single input to each tool module. The `state` is structured as:

>>> Agent.state = {
... 'config'            : {
...     <configuration variables>
... },
... 'prompt'            : <user input>,
... 'output'            : <streaming output of Agent>,
... 'memory'            : <agent memory>,
... 'process'           : {
...     'MODULE'        : <Tool module used to respond to user input>,
...     <module specific information>: {
...         ...
...     }
... },
... 'queue'             : [<list of instructions to follow>],
... 'queue pointer'     : <instruction pointer to the queue>,
... 'llm-api-calls'     : <number of LLM calls used by Agent>,
... 'recursion_depth'   : <amount of recursion the Agent is using>
>>> }
    
Class Methods
-------------

The `Agent` class is organized as follows:


"""

# Standard
import pandas as pd
import os
import shutil
from datetime import datetime as dt
import warnings
import json
import logging
import time
from typing import Optional, List
import pickle
import atexit

# Router
from semantic_router.layer import RouteLayer

# RAG
import chromadb

# LangChain
from langchain.vectorstores import Chroma
from langchain.embeddings import HuggingFaceEmbeddings
from langchain.chains import RetrievalQA
from langchain.llms import LlamaCpp
from langchain.callbacks.manager import CallbackManager
from langchain.callbacks.streaming_stdout import StreamingStdOutCallbackHandler
from langchain.prompts import PromptTemplate
from langchain.document_loaders import PyPDFLoader
from langchain.text_splitter import CharacterTextSplitter
from langchain.chains.question_answering import load_qa_chain
from langchain.output_parsers import CommaSeparatedListOutputParser
from langchain_nvidia_ai_endpoints import NVIDIAEmbeddings, ChatNVIDIA
from langchain.memory import ConversationBufferMemory

# LangChain Core
from langchain_core.callbacks.manager import CallbackManagerForLLMRun
from langchain_core.language_models.llms import LLM

# Library
from BRAD.planner import *
from BRAD.enrichr import *
from BRAD.scraper import *
from BRAD.router import *
from BRAD.rag import queryDocs, remove_repeats
from BRAD.gene_ontology import *
from BRAD.pythonCaller import *
from BRAD.llms import *
from BRAD.geneDatabaseCaller import geneDBRetriever
from BRAD.planner import planner
from BRAD.coder import code_caller
from BRAD.writer import summarizeSteps, chatReport
from BRAD import log
from BRAD.bradllm import BradLLM
from BRAD.constants import TOOL_MODULES

class Agent():
    """
    This class organizes the agentic capabilities of BRAD. It facilitates interactions 
    with external LLMs, tools, core modules, literature, and other databases while
    managing the chat state and history.
    
    Key functions include:
    
    1. **invoke(user_input)**: Responds to a single user input.
    2. **chat()**: Initiates an interactive session between the user and the BRAD agent.
    
    To address user queries, the agent employs semantic routing to select the appropriate tool module, generates responses using code from the chosen module, and tracks its state throughout the interaction.

    :param model_path: The path to the Llama model file, defaults to '/nfs/turbo/umms-indikar/shared/projects/RAG/models/llama-2-7b-chat.Q8_0.gguf'.
    :type model_path: str, optional
    :param persist_directory: The directory where the literature database is stored, defaults to "/nfs/turbo/umms-indikar/shared/projects/RAG/databases/Transcription-Factors-5-10-2024/".
    :type persist_directory: str, optional
    :param llm: The language model to be used. If None, it will be loaded within the function.
    :type llm: PreTrainedModel, optional
    :param ragvectordb: The RAG vector database to be used. If None, it will prompt the user to load it.
    :type ragvectordb: Chroma, optional
    :param embeddings_model: The embeddings model to be used. If None, it will be loaded within the function.
    :type embeddings_model: HuggingFaceEmbeddings, optional
    :param max_api_calls: The maximum number of api / llm calls BRAD can make
    :type max_api_calls: int, optional
    :param tools: The set of available tool modules. If None, all modules are available for use
    :type tools: list, optional
    :param gui: Indicates if the Agent is used in the GUI
    :type gui: boolean, optional

    :raises FileNotFoundError: If the specified model or database directories do not exist.
    :raises json.JSONDecodeError: If the configuration file contains invalid JSON.
    :raises KeyError: If required keys are missing from the configuration or chat status.
    """

    
    def __init__(self,
        model_path = '/nfs/turbo/umms-indikar/shared/projects/RAG/models/llama-2-7b-chat.Q8_0.gguf',
        persist_directory = "/nfs/turbo/umms-indikar/shared/projects/RAG/databases/DigitalLibrary-10-June-2024/",
        llm=None,
        ragvectordb=None,
        embeddings_model=None,
        restart=None,
        start_path=None,
        tools=None,
        name='BRAD',
        max_api_calls=None, # This prevents BRAD from finding infinite loops and using all your API credits,
        interactive=True,   # This indicates if BRAD is in interactive more or not
        config=None,        # This parameter lets a user specify an additional configuration file that will
                            # overwrite configurations with the same key
        gui=False
    ):
        # Auth: Joshua Pickard
        #       jpic@umich.edu
        # Date: June 4, 2024

        # Dev. Comments:
        # -------------------
        # This function initializes a chat session and the state variable
        #
        # History:
        # - 2024-06-04: wrote 1st draft of this code in the brad.chat() method
        # - 2024-07-10: refactored brad.py file to a class and converted the code
        #               used to initialize the chat session to initialize this class
        # - 2024-07-23: added interactive and max_api_call arguments
        # - 2024-07-29: added config (optional) argument to overwrite the defaults
        # - 2024-10-06: .chatstatus was renamed .state
        # - 2024-10-16: if the restart location doesn't have a log, then a new agent
        #               is created
        # - 2024-11-18: configurations are automatically read/written from created sessions
        #               when turning the Agent back on
        #
        # Issues:
        # - We should change the structure of the classes/modules. In this 1st iteration
        #   state was packed as a class variable and used similar to before, but it
        #   is likely reasonable to split this variable up into separate class variables.
        # - Current configurations only allow either a specific configuration file or a
        #   configuration file from a previous chat session (dominate) - JP

        self.state = self.load_state(config=config)
        self.name       = name.strip()
        self.state['interactive'] = interactive # By default a chatbot is not interactive
        self.state['gui'] = gui
        logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
    
        base_dir = os.path.expanduser('~')
        log_dir = os.path.join(base_dir, self.state['config']['log_path'])
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)

        # Check if it will be possible to restart the old session
        if restart is not None:
            log_path = os.path.join(restart, 'log.json')
            if not os.path.exists(log_path):
                restart = None

        if restart is None:
            # making this more human readable
            new_dir_name = start_path if start_path else dt.now().strftime("%B %d, %Y at %I:%M:%S %p")
            
            new_log_dir = os.path.join(log_dir, new_dir_name)
            try:
                os.makedirs(new_log_dir, exist_ok=True)
            except Exception as e:
                print(f"Failed to create directory '{new_log_dir}'. Error: {e}")
                fallback_dir = "C:\\Users\\jpic\\Documents\\BRAD-logs"  # Replace this with a directory that must exist
                print(f"Using fallback directory: {fallback_dir}")
                new_log_dir = fallback_dir
                os.makedirs(new_log_dir, exist_ok=True)  # This should always succeed since the fallback exists
            self.chatname = os.path.join(new_log_dir, 'log.json')
            self.chatlog  = {}
        else:
            new_log_dir = restart
            self.chatname = os.path.join(restart, 'log.json')
            self.chatlog  = json.load(open(self.chatname))

            state_file = os.path.join(new_log_dir, '.agent-state.pkl')
            
            if os.path.exists(state_file):
                try:
                    with open(state_file, 'rb') as f:
                        self.state = pickle.load(f)
                    logging.info(f"Loaded agent state from {state_file}")
                except Exception as e:
                    logging.error(f"Failed to load agent state from {state_file}: {e}")
            else:
                logging.info(f"No existing state file found in {new_log_dir}, starting fresh.")

            # Update from the old configurations
            config_file = os.path.join(new_log_dir, 'config.json')
            if os.path.exists(config_file):
                try:
                    saved_configs = self.load_config(config_file)
                    self.state['config'] = saved_configs
                    logging.info(f"Loaded configuration from {config_file}")
                except Exception as e:
                    logging.error(f"Failed to load configuration from {config_file}: {e}")
            else:
                logging.info(f"No configuration file found at {config_file}.")

        if max_api_calls is None:
            max_api_calls = 1000
        self.max_api_calls = max_api_calls
    
        # Initialize the dictionaries of tables and databases accessible to BRAD
        databases = {} # a dictionary to look up databases
        # tables = {}    # a dictionary to look up tables
    
        # Initialize the RAG database
        if llm is None:

            # By devault we use OpenAI
            llm = load_openai()

        if ragvectordb is None:
            if self.state['interactive']:
                state = log.userOutput('\nWould you like to use a database with ' + self.name + ' [Y/N]?', state=self.state)
                loadDB = input().strip().upper()
                if loadDB == 'Y':
                    ragvectordb, embeddings_model = self.load_literature_db(persist_directory) # load the literature database
                else:
                    ragvectordb, embeddings_model = None, None
            else:
                ragvectordb, embeddings_model = None, None
        
        databases['RAG'] = ragvectordb
        memory = ConversationBufferMemory(ai_prefix="BRAD")
    
        # Initialize the routers from router.py
        self.router = getRouter(available=tools)
    
        # Add other information to state
        # Assign values only if the key does not exist or is None/empty
        if 'llm' not in self.state or not self.state['llm']:
            self.state['llm'] = llm

        if 'memory' not in self.state or not self.state['memory']:
            self.state['memory'] = memory

        if 'databases' not in self.state or not self.state['databases']:
            self.state['databases'] = databases

        if 'output-directory' not in self.state or not self.state['output-directory']:
            self.state['output-directory'] = new_log_dir
    
        # Initialize all modules
        self.module_functions = self.getModules()
    
        # Start loop
        # only log if chat bot is fresh
        if restart is None:
            self.state = log.userOutput('Welcome to BRAD! The output from this conversation will be saved to ' + self.chatname + '. How can I help?', state=self.state)

            # Write an empty chat log
            self.chatlog, self.state = log.logger(self.chatlog, self.state, self.chatname, elapsed_time=0)

        # Ensure that the save_state function is registered to run at program exit
        # atexit.register(self.save_state)


    def save_state(self):
        """
        Saves the agent state to a file named '.agent-state.pkl' in the output directory.
        This method is registered with atexit to ensure it is called when the program exits.
        """
        # Auth: Joshua Pickard
        #       jpic@umich.edu
        # Date: October 15, 2024

        output_directory = self.state['output-directory']
        if not os.path.exists(output_directory):
            os.makedirs(output_directory)

        state_file = os.path.join(output_directory, '.agent-state.pkl')

        try:
            # Save the log to the json log file
            self.state['prompt'] = None
            self.state['output'] = None
            self.state['process'] = {
                'module' : 'SLEEP'
            }

            # Set the databases to none
            self.state['databases']['RAG'] = None
            self.state['llm'] = None

            # Save the state to a pickle file
            with open(state_file, 'wb') as f:
                pickle.dump(self.state, f)
            logging.info(f"Agent state saved to {state_file}")

            self.chatlog, self.state = log.logger(self.chatlog, self.state, self.chatname, elapsed_time=0)
            logging.info(f"Agent log written for power off")

        except Exception as e:
            logging.error(f"Failed to save agent state: {e}")


    def invoke(self, query):
        """
        Executes a single query using the chatbot, similar to invoking a language model.
    
        This method processes the user input, determines the appropriate routing 
        based on explicit commands or the content of the query, and generates a 
        response using the selected module. It also manages the state of the 
        chatbot throughout the execution.
    
        :param query: The user input to be processed by the chatbot.
        :type query: str
    
        :return: The output generated by the chatbot in response to the input query.
        :rtype: str
    
        :raises Exception: If an error occurs during the execution of the selected module.
    
        :example:
            >>> response = agent.invoke("What's the weather today?")
        
        :note: 
            Special commands recognized include:
            - "exit", "quit", "q", "bye": Ends the session.
            - "help": Displays help information.
            - "/set": Configures settings.
            - "/force": Forces the use of a specified routing function.
    
        This method logs the process and clears memory based on configuration settings.
        """
        # Auth: Joshua Pickard
        #       jpic@umich.edu
        # Date: June 4, 2024

        # Dev. Comments:
        # -------------------
        # This function executes a single user prompt with BRAD
        #
        # History:
        # - 2024-06-04: wrote 1st draft of this code in the brad.chat() method
        # - 2024-07-10: refactored brad.py file to a class and converted the code
        #               used to execute a single prompt into the invoke method

        # start the clock
        start_time = time.time()
        
        # This line packs a query into the state variable we were using previously
        self.state['prompt'] = query
        
        # Handle explicit commands and routing
        if self.state['prompt'].lower() in ['exit', 'quit', 'q', 'bye']:         # check to exit
            return False
        elif self.state['prompt'].lower() == 'help':              # print man to the screen
            self.chatbotHelp()
            return True
            # continue
        # Routing
        elif self.state['prompt'].startswith('/set'):             # set a configuration variable
            self.state = self.reconfig()
            return True
            # continue
        # Continue previous module
        elif self.state.get('continue-module') is not None:
            route = self.state['continue-module'][0]

        # Use router to select correct module
        elif '/force' not in self.state['prompt'].split(' '):     # use the router
            route = self.router(self.state['prompt']).name
            if route is None:
                route = 'RAG'
        # Forced routes
        else:
            route = self.state['prompt'].split(' ')[1]            # use the forced router
            if self.state['config']['ROUTER']['build router db']:
                buildRoutes(self.state['prompt'])
            self.state['prompt'] = " ".join(self.state['prompt'].split(' ')[2:]).strip()

        # Outputs
        self.state = log.userOutput('==================================================', state=self.state)
        self.state = log.userOutput(self.name + ' >> ' + str(len(self.chatlog)) + ': ', state=self.state)

        # select appropriate routeing function
        if route.upper() in self.module_functions.keys():
            routeName = route.upper()
        else:
            routeName = 'RAG'

        log.debugLog(routeName, state=self.state)

        # get the specific module to use
        module = self.module_functions[routeName]

        # Standardize state/logging schema for all modules
        #     see: https://github.com/Jpickard1/RAG-DEV/wiki/Logging
        self.state['process'] = {'module' : routeName,
                                 'steps'  : []
                            }
        self.state['output'] = ""

        # get current output files
        output_files = utils.outputFiles(self.state)
        IP = self.state['queue pointer']
        # if not self.state['interactive']:
        #     self.state['queue'] = [[]] # don't let it us an empty queue if it is not interactive
        
        # Query module
        try:
            self.state = module(self.state)
        except:
            log.debugLog('An error occurred while using a tool.', state=self.state)

        # Remove the item that was executed. We need must do it after running it for the current file naming system.
        log.debugLog('\n\n\nroute\n\n\n', state=self.state)
        log.debugLog(route, state=self.state)
        if len(self.state['queue']) != self.state['queue pointer'] and route != 'PLANNER':
            # log.debugLog(self.state['queue'], state=self.state)
            new_output_files = utils.outputFiles(self.state)
            new_output_files = list(set(new_output_files).difference(set(output_files)))
            if 'output' not in self.state['queue'][IP].keys():
                self.state['queue'][IP]['outputs'] = new_output_files
            self.state = utils.makeNamesConsistent(self.state, new_output_files)
            if self.state['process']['module'] != 'ROUTER':
                self.state['queue pointer'] += 1

        # Clear memory
        if self.state['config']['forgetful']:
            self.state['memory'].clear()

        # stop the clock
        end_time = time.time()
        elapsed_time = end_time - start_time
        
        # Log and reset these values
        self.chatlog, self.state = log.logger(
            self.chatlog,
            self.state,
            self.chatname,
            elapsed_time=elapsed_time
        )
        return self.state['output']

    def chat(self):
        """
        Opens an interactive chat session where users can execute a series of prompts.
    
        This method allows users to engage in a back-and-forth dialogue with the chatbot. 
        Users can input queries, which the chatbot processes and responds to until 
        the session is terminated. It supports both direct user input and queued prompts.
    
        The chat session maintains a record of the conversation and tracks the number 
        of API calls made to language models. It ensures that the session can be exited 
        gracefully and provides feedback about the conversation's context.
    
        :example:
            >>> agent.chat()
        
        :note:
            The session continues until the user explicitly decides to exit by 
            inputting commands like "exit", "quit", or "bye".
    
            If a queue of prompts is available, the chatbot will process 
            them in sequence rather than waiting for user input.
    
            The memory can be temporarily integrated to enrich queries, 
            but its management is handled with care to avoid unintended 
            modifications.
        """
        # Auth: Joshua Pickard
        #       jpic@umich.edu
        # Date: June 4, 2024

        # Dev. Comments:
        # -------------------
        # This function executes a series of prompts in an interactive mannor with BRAD
        #
        # History:
        # - 2024-06-04: wrote 1st draft of this code in the brad.chat() method
        # - 2024-07-10: refactored brad.py file to a class and converted the code
        #               used to execute consecutive prompts into this function
        # - 2024-07-21: added llm-api-calls to chat status to prevent the rerouting/
        #               planner modules from executing unnecessarily long loops.
        self.state['interactive'] = True
        while True:
            # Begin processing a new prompt
            print('==================================================')

            # By default the memory is not messed with in BRAD (but on special occasions it is!)
            resetMemory = False

            # Get the prompt from the user or from the queue
            if len(self.state['queue']) != 0 and self.state['queue pointer'] < len(self.state['queue']) and self.state['queue pointer'] != 0 and not self.state['interactive']:
                query = self.state['queue'][self.state['queue pointer']]['prompt']

                # update memory to use previous points of the pipeline
                if 'inputs' in list(self.state['queue'][self.state['queue pointer']].keys()):
                    print(query)
                    print(self.state['queue'][self.state['queue pointer']]['module'])
                    if self.state['queue'][self.state['queue pointer']]['module'] == 'RAG':
                        print('RAG MODULE')
                        print(f"self.state['queue'].keys()={self.state['queue'].keys()}")
                        inputStages = self.state['queue'][self.state['queue pointer']]['inputs']
                        print(inputStages)
                        query += "**Previous Work**\nYou have generated the following data previously in the pipeline:\n\n"
                        for ISIP in inputStages:
                            fileName  = self.state['queue'][ISIP]['outputs']
                            print(fileName)
                            for file in fileName:
                                file = os.path.join(self.state['output-directory'], file)
                                if not os.path.exists(file):
                                    continue
                                try:
                                    print(file)
                                    dfMemory  = pd.read_csv(file)
                                    if 'state' in df.columns:
                                        query += "Top Ranked Biomarkers:\n"
                                        query += str(df['state'].values[:100])[1:-1]
                                    if 'p_val' in df.columns:
                                        query += "Relevant Biological Processes from Enrichr:"
                                        query += str(df[['path_name']].values[:10])
                                except:
                                    print('this part of the code doesnt work so great, yet!')
            else:
                query = input('Input >> ')  # get query from user
            
            # This line needs to change to check if we should exist the chat session
            output = self.invoke(query)

            # output is false on exit
            if not output:
                break

            # reset memory
            # if resetMemory:
            #    self.resetMemory()
                
            # update llm-api-calls
            newCalls = self.getLLMcalls(self.state['process']['steps'])
            self.state['llm-api-calls'] += newCalls
            log.debugLog(f'current llm calls={self.state["llm-api-calls"]}', state=self.state)
            if self.state['llm-api-calls'] > self.max_api_calls:
                log.debugLog('The maximum number of llm calls has been exceeded', state=self.state)
                break
        self.state['interactive'] = False
        self.state = log.userOutput("Thanks for chatting today! I hope to talk soon, and don't forget that a record of this conversation is available at: " + self.chatname, state=self.state)

    def get_display(self):
        """
        This function returns the history of all inputs/outputs to the agent. This is intended
        for use by the GUI, as it will allow the user to jump between sessions while loading in
        the history of the old session.

        :returns: a list of strings
        :rtype: list
        """
        log.debugLog("Get Display Method", display=True)
        # numIOpairs = len(self.chatlog.keys())
        display = []
        log.debugLog("display = []", display=True)
        for i in self.chatlog.keys():
            if 'module' in self.chatlog[i]['process'].keys() and self.chatlog[i]['process']['module'] == 'SLEEP':
                continue
            elif 'MODULE' in self.chatlog[i]['process'].keys() and self.chatlog[i]['process']['MODULE'] == 'SLEEP':
                continue
            else:
                display.append((self.chatlog[i]['prompt'], None))
                display.append((self.chatlog[i]['output'], self.chatlog[i]))
        log.debugLog(f"{display=}", display=True)
        return display

    @property
    def llm(self):
        """Get the current LLM."""
        return self.state['llm']

    def set_llm(self, llm):
        """Set the LLM and handle any related logic."""
        if not llm:
            raise ValueError("LLM cannot be None")
        
        self.state['llm'] = llm

    def updateMemory(self):
        """
        Thils function lets BRAD reset his memory to focus on specific previous interactions. This is useful
        when BRAD is executing a pipeline and want to manage how input flows from one section to the next.

        .. warning::
            This function may be removed in the near future.
            
        """
        # Auth: Joshua Pickard
        #       jpic@umich.edu
        # Date: July 28, 2024

        log.debugLog('updateMemory is starting', state=state)
        
        # get previous stages that must occur in the input
        inputs = self.state['queue'][self.state['queue pointer']]['input']

        # build a new memory object based upon the previous inputs and outputs of each stage
        memory = ConversationBufferMemory()
        for stage in inputs:
            stageHistory = self.state['queue'][stage]['history']
            memory.add_user_message(stageHistory['input'][-1])
            memory.add_ai_message(stageHistory['output'][-1])

        # Update the state main memory
        state['main memory'] = state['memory']
        state['memory'] = memory
        return
                

    def resetMemory(self):
        """
        Resets the state of the memory by restoring the main memory from the recent messages.
    
        This function undoes the effects of the `updateMemory` method by taking
        the most recent messages from the stage memory and adding them back 
        into the main memory. It is designed to help maintain a consistent 
        memory state throughout the agent's execution.
    
        .. warning::
            This function may be removed in the near future.
    
        :return: None
        :rtype: None
    
        :example:
            >>> agent.resetMemory()
        """
        # Auth: Joshua Pickard
        #       jpic@umich.edu
        # Date: July 28, 2024

        # Update the state main memory
        mainMemory = state['main memory']
        stageMemory = state['main memory']
        
        recent_messages = stageMemory.buffer[-2:]  # Adjust -2 index to control # of recent messages
        for msg in recent_messages:
            mainMemory.add_message(msg['role'], msg['content'])

        state['memory'] = mainMemory
        return
    
    def getLLMcalls(self, steps):
        """
        Counts the number of times the LLM has been called used the agent's execution.
        
        :param steps: A list of steps from the agents log that have been executed
        :type steps: list
        
        :return: The total number of LLM calls made by the agent.
        :rtype: int
        
        :example:
            >>> num_calls = agent.getLLMcalls(steps)
        """
        # Auth: Joshua Pickard
        #       jpic@umich.edu
        # Date: July 2024
        newLLMcalls = 0
        emptyLLMlog = log.llmCallLog()
        for step in steps:
            if all(k in step.keys() for k in emptyLLMlog.keys()):
                newLLMcalls += 1
        return newLLMcalls
    
    def getModules(self):
        """
        Returns a dictionary mapping module names to their corresponding function handles for various tasks.
    
        :param None: This function does not take any parameters.
    
        :raises None: This function does not raise any specific errors.
    
        :return: A dictionary where the keys are module names and the values are function handles for tasks such as querying Enrichr, web scraping, generating seaborn plots, querying documents, and calling Snakemake.
        :rtype: dict
    
        """
        # Auth: Joshua Pickard
        #       jpic@umich.edu
        # Date: June 4, 2024
        
        # this + imports should be the only code someone needs to write to add a new module
        module_functions = {
            'DATABASE'   : geneDBRetriever,     # gget
            'SCRAPE' : webScraping,      # webscrapping
            'RAG'    : queryDocs,        # standard rag
            'PYTHON' : callPython,
            'PLANNER': planner,
            'CODE'   : code_caller,
            'WRITE'  : chatReport, # summarizeSteps,
            'ROUTER' : reroute,
        }
        return module_functions

    def load_config(self, configfile=None):
        """
        Loads the `Agent` configuration settings from a JSON file.
    
        :param None: This function does not take any parameters.
    
        :raises FileNotFoundError: If the configuration file is not found.
        :raises json.JSONDecodeError: If the configuration file contains invalid JSON.
    
        :return: A dictionary containing the configuration settings.
        :rtype: dict
        """
        # Auth: Joshua Pickard
        #       jpic@umich.edu
        # Date: June 4, 2024

        # History:
        # - 2024-06-04: wrote 1st draft of this code
        # - 2024-07-29: added a new argument to allow additional configuration
        #               files to overwrite the defaults

        def deep_update(original, updates):
            """
            Recursively update the original dictionary with the updates dictionary.
            """
            for key, value in updates.items():
                if isinstance(value, dict) and key in original:
                    original[key] = deep_update(original.get(key, {}), value)
                else:
                    original[key] = value
            return original
        
        # Read the default configurations
        current_script_path = os.path.abspath(__file__)
        current_script_dir = os.path.dirname(current_script_path)
        file_path = os.path.join(current_script_dir, 'config', 'config.json')
        with open(file_path, 'r') as f:
            defaultConfigs = json.load(f)

        # print(defaultConfigs)

        # Read the specific configurations
        if configfile:
            with open(configfile, 'r') as f:
                newConfigs = json.load(f)
            defaultConfigs = deep_update(defaultConfigs, newConfigs)

        # print(defaultConfigs)
        return defaultConfigs    
    
    def save_config(self):
        """
        Saves the agent configuration settings to a JSON file.
    
        :param config: A dictionary containing the configuration settings to be saved.
        :type config: dict
    
        :raises FileNotFoundError: If the directory for the configuration file is not found.
        :raises TypeError: If the configuration dictionary contains non-serializable values.    
        """
        # Dev. Comments
        # History:
        # - 2024-11-18: This method modifies the objects configurations but does not change
        #               the configurations of the package. (JP)
        #
        # Auth: Joshua Pickard
        #       jpic@umich.edu
        # Date: June 4, 2024

        # current_script_path = os.path.abspath(__file__)
        # current_script_dir = os.path.dirname(current_script_path)
        # file_path = os.path.join(current_script_dir, 'config', 'config.json')
        file_path = os.path.join(self.chatname[:-8], 'config.json')
        print(f"{file_path=}")
        with open(file_path, 'w') as f:
            json.dump(self.state['config'], f, indent=4)

    def reconfig(self):
        """
        Updates a specific configuration setting based on the given chat status and saves the updated configuration.
    
        :param chat_status: A dictionary containing the current chat status, including the prompt and configuration.
        :type chat_status: dict
    
        :raises KeyError: If the specified configuration key is not found in the chat status.
        :raises ValueError: If the value cannot be converted to an integer or float when applicable.
    
        :return: The updated chat status dictionary.
        :rtype: dict
    
        """
        # Auth: Joshua Pickard
        #       jpic@umich.edu
        # Date: June 4, 2024
        
        prompt = self.state['prompt']
        _, key, value = prompt.split(maxsplit=2)
        if key == 'debug':
            value = (value.lower() == 'true')
        try:
            value = int(value)
        except ValueError:
            try:
                value = float(value)
            except ValueError:
                value = str(value)
        if key in self.state['config']:
            state['config'][key] = value
            self.save_config()
            state = log.userOutput("Configuration " + str(key) + " updated to " + str(value), state=self.state)
        else:
            state = log.userOutput("Configuration " + str(key) + " not found", state=self.state)
        return state

    def load_state(self, config=None):
        """
        Initializes and loads the agent state with default values and configuration settings.
    
        :param None: This function does not take any parameters.
    
        :raises FileNotFoundError: If the configuration file is not found.
        :raises json.JSONDecodeError: If the configuration file contains invalid JSON.
    
        :return: A dictionary representing the chat status with initial values and loaded configuration.
        :rtype: dict
        """
        # Auth: Joshua Pickard
        #       jpic@umich.edu
        # Date: June 4, 2024
    
        state = {
            'config'            : self.load_config(configfile=config),
            'prompt'            : None,
            'output'            : None,
            'process'           : {},
            'current table'     : {'key':None, 'tab':None},
            'current documents' : None,
            'tables'            : {},
            'documents'         : {},
            'plottingParams'    : {},
            'matlabEng'         : None,
            'experiment'        : False,
            'queue'             : [],
            'queue pointer'     : 0,
            'llm-api-calls'     : 0,
            'search'            : {
                'used terms' : [],
            },
            'recursion_depth': 0,
            'continue-module': None, # None (if not continuing in a module) or tuple (with [0] being the module name)
            'gui':None,              # boolean to indicate if the agent is for the gui
            'interactive':False
        }
        return state

    def load_literature_db(
        self,
        persist_directory = "/home/acicalo/BRAD/data/RAG_Database",
        db_name = "DB_cosine_cSize_700_cOver_200"
    ):
        """
        Loads a literature database using specified embedding model and settings.
    
        :param persist_directory: The directory where the database is stored, defaults to "/nfs/turbo/umms-indikar/shared/projects/RAG/databases/Transcription-Factors-5-10-2024/"
        :type persist_directory: str, optional
    
        :raises FileNotFoundError: If the specified directory does not exist or is inaccessible.
        :raises Warning: If the loaded database contains no articles.
    
        :return: A tuple containing the vector database and the embeddings model.
        :rtype: tuple

        The `persist_directory` should point to a directory that has this structure:

        >>> [Oct 16 19:28]  persist_directory
        ... └── [Oct 16 19:28]  DB_cosine_cSize_700_cOver_200
        ...     ├── [Oct 16 19:28]  aaa2c989-0e39-4be8-82b4-139ae2784c00
        ...     │   ├── [Oct 16 19:28]  data_level0.bin
        ...     │   ├── [Oct 16 19:28]  header.bin
        ...     │   ├── [Oct 16 19:28]  length.bin
        ...     │   └── [Oct 16 19:28]  link_lists.bin
        >>>     └── [Oct 16 19:28]  chroma.sqlite3
    
        """
        # Dev. Comments
        # History:
        # - 2024-06-04: 1st version of this was written
        # - 2024-10-17: changes to the pathing were made

        # Auth: Joshua Pickard
        #       jpic@umich.edu
        # Date: June 4, 2024
    
        # load database
        embeddings_model = HuggingFaceEmbeddings(model_name='BAAI/bge-base-en-v1.5')        # Embedding model
        _client_settings = chromadb.PersistentClient(path=os.path.join(persist_directory, db_name))
        vectordb = Chroma(persist_directory=persist_directory, embedding_function=embeddings_model, client=_client_settings, collection_name=db_name)
        if len(vectordb.get()['ids']) == 0:
            print('The loaded database contains no articles. See the database: ' + str(persist_directory) + ' for details')
            warnings.warn('The loaded database contains no articles. See the database: ' + str(persist_directory) + ' for details')
        return vectordb, embeddings_model

    def chatbotHelp(self):
        """
        Displays a help message to the user with information about the BRAD agents's capabilities and special commands.
        """
        # Auth: Joshua Pickard
        #       jpic@umich.edu
        # Date: May 19, 2024
        help_message = """
        Welcome to our RAG chatbot Help!
        
        You can chat with the llama-2 llm and several scientifici databases including:
        - retrieval augmented generation
        - web scraping
        - search enrichr and gene ontology databases
        - run codes
        - and more!
    
        Special commands:
        /set   - Allows the user to change configuration variables.
        /force - Allows the user to specify which database to use.
    
        For example:
        /set config_variable_name new_value
        --force option_name
    
        Enjoy chatting with the chatbot!
        """
        self.state = log.userOutput(help_message, state=self.state)
        return

    def to_langchain(self):
        """
        This function constructs an object that may be used as an llm in langchain.

        :return: a LangChain compatible LLM instance
        :rtype: `BradLLM`
        """
        # Auth: Joshua Pickard
        #       jpic@umich.edu
        # Date: July 10, 2024
        llm = BradLLM(bot=self)
        return llm


class AgentFactory():
    """
    The AgentFactory mechanism allows us to instantiate, terminate and maintain 
    bot sessions from within the module. Removes the need to have global agents.
    The factory generates a default agent with default parameters if no input is
    given. Based on the session input given instantiates a new agent with that 
    particular session restored

    Provides decoupling of objects used from execution logic. If a new agent class 
    is implemented the get_agent function needs to be updated appropriately.

    Functions:
    1. **get_agent**: instantiates the actual agent and returns the particular agent based on initialization


    :param tools: The set of available tool modules. If None, all modules are available for use
    :type tools: list, optional
    :param session_path: The path to where the bot session is stored. If None, generates a new agent
    :type session_path: str, optional
    :param start_path: The path to where a new bot session can be started. If None, generates a new agent
    :type start_path: str, optional
    :param interactive: Sets BRAD's mode to interactive or non inteactive. Default mode is non Interactive
    :type tools: bool, optional
    """

    def __init__(self, tool_modules=TOOL_MODULES, session_path=None, start_path=None, interactive=False, db_name=None, persist_directory=None, llm_choice=None, gui=None, temperature=0):
        self.interactive = interactive
        suffix = '/log.json'
        if session_path and (session_path.endswith(suffix)):
            session_path = session_path[: -len(suffix)]
        self.session_path = session_path
        self.tool_modules = tool_modules
        self.start_path = start_path
        self.gui=gui
        if db_name:
            self.db_name = db_name
        else:
            self.db_name = None

        if persist_directory:
            self.persist_directory = persist_directory
        else:
            self.persist_directory = None

        self.llm_choice = llm_choice
        self.temperature = temperature
        

    def get_agent(self):
        """
        The agent function for instantiating a new agent or retrieve an existing agent
        """
        if self.session_path:
            agent = Agent(interactive=self.interactive, tools=self.tool_modules, restart=self.session_path, gui=self.gui)
        elif self.start_path:
            agent = Agent(interactive=self.interactive, tools=self.tool_modules, start_path=self.start_path, gui=self.gui)
        else:
            agent = Agent(interactive=self.interactive, tools=self.tool_modules, gui=self.gui)


        if self.db_name and self.persist_directory:
            db, _ = agent.load_literature_db(persist_directory=self.persist_directory, db_name=self.db_name)
            agent.state['databases']['RAG'] = db
        elif self.db_name == None and self.persist_directory==None:
            agent.state['databases']['RAG'] = None

        if self.llm_choice:
            llm = llm_switcher(
                self.llm_choice,
                temperature = self.temperature,
            )
            agent.set_llm(llm)
        return agent


"""
This module manages routing decisions for both individual user inputs and agentic workflows using the semantic_router 
library.

Scope
-----

It serves two key purposes within the BRAD framework: determining which tool or module to use for a specific 
input, and orchestrating multi-step workflows by selecting the next stage in the process.

1. **Tool Selection for Single Inputs**:  
   When receiving input from a user, the router evaluates the context and selects the appropriate tool or module to handle 
   the input. This ensures that the most suitable functionality is used to process each user request, improving the efficiency 
   and accuracy of interactions. Semantic routing is used for this level or organization, and the routing database may be dynamically
   updated as more user queries are recieved.

2. **Agentic Workflow Management**:  
   For agent-driven processes, the router manages the progression of tasks in a predefined workflow designed by the `planner`. It selects the next 
   step based on the current state and the broader workflow plan, ensuring smooth transitions between stages. This feature is organized by the LLM and
   is essential for workflows that involve multiple steps, tools, or decision points.

Available Methods
-----------------

This module contains the following methods:

"""

import os
import json
from semantic_router import Route
from semantic_router.layer import RouteLayer
from semantic_router.encoders import HuggingFaceEncoder

from langchain.prompts import PromptTemplate
from langchain_core.prompts.prompt import PromptTemplate

from BRAD.promptTemplates import rerouteTemplate
from BRAD import log
from BRAD import utils

def reroute(state):
    """
    Reroutes the conversation flow for agentic workflows based on the current queue pointer and the user prompt.
    
    This method uses the agent history and ongoing conversation, along with an LLM, 
    to determine the subsequent step in the agentic workflow designed by the planner.
    
    :param state: A dictionary that holds the current state of the conversation, including the language model, 
                       the user prompt, the process queue, and any other relevant information necessary for 
                       effectively rerouting the conversation.
    
    :returns: An updated state dictionary reflecting changes made during the rerouting process.
    :rtype: dict
    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: June 24, 2024
    log.debugLog('Call to REROUTER', state=state)
    llm = state['llm']
    prompt = state['prompt']
    queue  = state['queue']

    # Build chat history
    chatlog = json.load(open(os.path.join(state['output-directory'], 'log.json')))
    history = ""
    for i in chatlog.keys():
        history += "========================================"
        history += '\n'
        history += "Input: "  + chatlog[i]['prompt']  + r'\n\n'
        history += "Output: " + chatlog[i]['output'] + r'\n\n'
    log.debugLog(history, state=state)
    
    # Put history into the conversation
    template = rerouteTemplate()
    template = template.format(chathistory=history,
                               step_number=state['queue pointer'])
    PROMPT = PromptTemplate(input_variables=["user_query"], template=template)
    chain = PROMPT | state['llm']
    res = chain.invoke(prompt)
    
    # Extract output
    log.debugLog(res, state=state)
    log.debugLog(res.content, state=state)
    # nextStep = int(res.content.split('=')[1].split('\n')[0].strip())
    nextStep = utils.find_integer_in_string(res.content.split('\n')[0])
    log.debugLog('EXTRACTED NEXT STEP', state=state)
    log.debugLog('Next Step=' + str(nextStep), state=state)
    log.debugLog(f'(nextStep is None)={(nextStep is None)}', state=state)
    if nextStep is None:
        nextStep = state['queue pointer'] + 1
    if str(nextStep) not in prompt:
        log.debugLog(f'the next step identified was not valid according to the rerouting instructions. As a result, state["queue pointer"]={state["queue pointer"]+1}', state=state)
        nextStep = state['queue pointer'] + 1
    state['process']['steps'].append(log.llmCallLog(
        llm     = llm,
        prompt  = template,
        input   = prompt,
        output  = res.content,
        parsedOutput = {'next step': nextStep},
        purpose = "Route to next step in pipeline"
    ))
    
    # Modify the queued prompts
    state['queue pointer'] = nextStep
    return state

def read_prompts(file_path):
    """
    Reads a text file where each line contains a sentence and returns a list of non-empty sentences. The
    files contain previously used user inputs associated with tool modules.

    This function opens the specified text file, reads each line, and returns a list containing the 
    sentences. Leading and trailing whitespace is removed from each line, and empty lines are ignored.

    :param file_path: The path to the text file to be read.
    :type file_path: str

    :raises FileNotFoundError: If the specified file cannot be found.

    :return: A list of non-empty sentences extracted from the text file.
    :rtype: list[str]
    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: May 19, 2024
    sentences = []
    with open(file_path, 'r') as file:
        for line in file:
            # Strip any leading/trailing whitespace characters (including newline)
            sentence = line.strip()
            if sentence:  # Avoid adding empty lines
                sentences.append(sentence)
    return sentences

def add_sentence(file_path, sentence):
    """
    Adds a new sentence to the specified text file. These files contain user inputs associated with tool modules.

    :param file_path: The path to the text file where the sentence is to be added.
    :type file_path: str
    :param sentence: The sentence to be added to the text file.
    :type sentence: str

    :raises FileNotFoundError: If the specified file does not exist or cannot be created.

    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: May 19, 2024
    with open(file_path, 'a') as file:
        file.write(sentence.strip() + '\n')

def getRouterPath(file):
    """
    Constructs and returns the absolute path to a file located in the 'routers' directory.

    This function determines the current script's directory and constructs the absolute path 
    to a specified file within the 'routers' subdirectory.

    :param file: The name of the file whose path is to be constructed.
    :type file: str

    :return: The absolute path to the specified file in the 'routers' directory.
    :rtype: str
    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: June 7, 2024
    current_script_path = os.path.abspath(__file__)
    current_script_dir = os.path.dirname(current_script_path)
    file_path = os.path.join(current_script_dir, 'routers', file) #'enrichr.txt')
    return file_path
    
def getRouter(available=None):
    """
    Constructs a semantic router layer configured to determine the correct tool module for user queries.
    This routing layer uses the routing files and previously used user inputs as data to predict which
    tool module to use. These routing files are updated over time.

    :param available: list of avaialble modules. If available is `None`, then all modules are available by default.
    :type available: list, optional

    :return: A router layer configured with predefined routes for tasks such as querying Enrichr, web scraping, and generating tables.
    :rtype: RouteLayer

    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: May 16, 2024

    # Dev. Comments:
    # -------------------
    #
    # History:
    # - 2024-05-16: first version of this method
    # - 2024-10-15: this method is modified to allow a user to constrict
    #               the set of modules available to an agent.

    routes = []

    # Digital Library
    if available is None or 'GGET' in available:
        routeGget = Route(
            name = 'GGET',
            utterances = read_prompts(getRouterPath('enrichr.txt'))
        )
        routes.append(routeGget)
    if available is None or 'SCRAPE' in available:
        routeScrape = Route(
            name = 'SCRAPE',
            utterances = read_prompts(getRouterPath('scrape.txt'))
        )
        routes.append(routeScrape)

    # Lab Notebook
    if available is None or 'RAG' in available:
        routeRAG = Route(
            name = 'RAG',
            utterances = read_prompts(getRouterPath('rag.txt'))
        )
        routes.append(routeRAG)

    # SOFTWARE
    if available is None or 'CODE' in available:
        routeCode = Route(
            name = 'CODE',
            utterances = read_prompts(getRouterPath('code.txt'))
        )
        routes.append(routeCode)
    if available is None or 'PYTHON' in available:
        routePython = Route(
            name = 'PYTHON',
            utterances = read_prompts(getRouterPath('python.txt'))
        )
        routes.append(routePython)

    # Additional Core Modules
    if available is None or 'PLANNER' in available:
        routePlanner = Route(
            name = 'PLANNER',
            utterances = read_prompts(getRouterPath('planner.txt'))
        )
        routes.append(routePlanner)
    if available is None or 'WRITE' in available:
        routeWrite = Route(
            name = 'WRITE',
            utterances = read_prompts(getRouterPath('write.txt'))
        )
        routes.append(routeWrite)
    if available is None or 'ROUTER' in available:
        routeRoute = Route(
            name = 'ROUTER',
            utterances = read_prompts(getRouterPath('router.txt'))
        )
        routes.append(routeRoute)

    # Encoder to embed the routeing examples
    encoder = HuggingFaceEncoder(device='cpu')

    # Construct route layer
    router = RouteLayer(encoder=encoder, routes=routes)    

    return router

def buildRoutes(prompt):
    """
    Constructs routes based on the provided prompt and updates the corresponding text files with the new prompts.

    This function processes the input prompt to identify specific commands (e.g., `/force`) 
    and builds a new prompt that is then appended to the designated router text file based on 
    the identified route. Each command maps to a specific file, and if the command is not 
    recognized, a KeyError will be raised.

    :param prompt: The prompt containing the information to be added to the router.
    :type prompt: str
    :raises KeyError: If the specified route is not found in the paths dictionary.

    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: May 19, 2024
    words = prompt.split(' ')
    rebuiltPrompt = ''
    i = 0
    while i < (len(words)):
        if words[i] == '/force':
            route = words[i + 1].upper()
            i += 1
        else:
            rebuiltPrompt += (' ' + words[i])
        i += 1
    paths = {
        'DATABASE': getRouterPath('enrichr.txt'),
        'SCRAPE'  : getRouterPath('scrape.txt'),
        'RAG'     : getRouterPath('rag.txt'),
        'TABLE'   : getRouterPath('table.txt'),
        'DATA'    : getRouterPath('data.txt'),
        'SNS'     : getRouterPath('sns.txt'),
        'MATLAB'  : getRouterPath('matlab.txt'),
        'PYTHON'  : getRouterPath('python.txt'),
        'PLANNER' : getRouterPath('planner.txt'),
        'CODE'    : getRouterPath('code.txt'),
        'WRITE'   : getRouterPath('write.txt'),
        'ROUTER'  : getRouterPath('router.txt')
    }
    filepath = paths[route]
    add_sentence(filepath, rebuiltPrompt)




"""
Module: writer

This module is responsible for compiling chat reports, histories, and logs from various interactions with the 
large language model (LLM) and generating a comprehensive PDF document. The module streamlines the process of 
documenting interactions, making it easier for users to review and analyze conversations.

Usage:
------
This module is intended to be used in scenarios where there is a need to document interactions with the LLM. 
By converting chat histories and logs into a well-organized PDF file, users can maintain a clear record of 
conversations, facilitating further analysis and reflection on the interactions.
"""
import pandas as pd
import os
import re
import json
from tabulate import tabulate
from datetime import datetime

from langchain import PromptTemplate, LLMChain
from langchain.output_parsers.json import SimpleJsonOutputParser
from langchain.output_parsers import PydanticOutputParser
from langchain_core.prompts import PromptTemplate
# from langchain_core.pydantic_v1 import BaseModel, Field

from pydantic import BaseModel, Field


from BRAD import log
from BRAD import utils
from BRAD.promptTemplates import setReportTitleTemplate, summarizeAnalysisPipelineTemplate, summarizeDatabaseCallerTemplate, summarizeRAGTemplate
from BRAD.pythonCaller import find_py_files, get_py_description, read_python_docstrings, pythonPromptTemplate, extract_python_code, execute_python_code

def chatReport(state):
    """
    Generates a report of the chat conversation based on the logged interactions 
    and writes it to a PDF file using a LaTeX template.

    Args:
        state (dict): A dictionary containing the language model, user prompt, 
                           output directory, and other relevant information for generating 
                           the chat report.

    Returns:
        dict: The updated state containing the path to the generated PDF report and 
              any updates made during the report generation process.

    Example
    -------
    >>> state = {
    ...     'llm': llm_instance,
    ...     'prompt': "Generate chat report",
    ...     'output-directory': '/path/to/output',
    ... }
    >>> updated_state = chatReport(state)
    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: July 2, 2024
    log.debugLog("Chat Report Called", state=state)
    query    = state['prompt']
    llm      = state['llm']

    # TODO: this should be path dependent in the configurations
    # Select the latex template. The specific latex template is hardcoded for now, but a latex-template-path
    # indicator has been added to the config.json file and we can do a selection process similar to how scripts
    # are selected to run
    latexTemplate = getLatexTemplate('/home/jpic/latex-test/templates/dmdBiomarkerEnrichment.tex')
    
    # Read the full chat history
    chatlog = json.load(open(os.path.join(state['output-directory'], 'log.json')))
    chathistory = getChatInputOutputs(state, chatlog)
    log.debugLog(chathistory, state=state)

    # Get fields that are always set
    title = getReportTitle(chathistory, state=state)
    log.debugLog("title= " + str(title), state = state)
    date  = getReportDate()
    log.debugLog("date = " + str(date) , state = state)

    # Fill in the latex template
    latexTemplate = latexTemplate.replace("BRAD-TITLE", title)
    latexTemplate = latexTemplate.replace("BRAD-DATE",  date)

    # Loop over possible keys we put in the latex templates:
    partsOfReport = ['SUMMARY', 'BODY']
    for part in partsOfReport:
        key = 'BRAD-' + part
        if key in latexTemplate:
            if key == 'BRAD-SUMMARY':
                section = getReportSummary(chathistory, state=state)
            if key == 'BRAD-BODY':
                section = getReportBody(state, chatlog)
            latexTemplate = latexTemplate.replace(key, section)

    # After filling in the report template write it to a pdf file
    state, report_file = reportToPdf(state, latexTemplate)
    state = utils.compile_latex_to_pdf(state, report_file)
    
    return state

def getReportDate():
    """
    Returns the current date formatted as a string in the format: "Day of the week, Month Day, Year".

    Returns:
        str: Formatted date string.
    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: July 2, 2024
    today = datetime.now()
    formatted_date = today.strftime("%A, %B %d, %Y")
    return formatted_date

def getReportTitle(chathistory, state):
    """
    Uses a language model to generate a descriptive title for a report summarizing the work done in the chatlog.

    Args:
        chathistory (dict): A dictionary containing the chat history with user inputs and model outputs.
        state (dict): A dictionary containing the current chat status, including the language model instance 
                           and user prompt.

    Returns:
        str: The generated report title based on the chatlog and user prompt.
    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: July 2, 2024
    llm      = state['llm']
    template = """Given the following chatlog, generate a descriptive title for a report summarizing the work done. Please keep the title concise. This report is in reposnse to the following user request:
{userinput}

**Chatlog Overview:**
{chatlog}
"""
    template = template.format(userinput=state['prompt'], chatlog=chathistory)
    log.debugLog(f"{template=}", state=state)
    prompt = PromptTemplate(template=template, input_variables=["chatlog"])
    json_prompt = PromptTemplate.from_template(
        "Return a JSON object with an `title` and `abstract` keys in response to the following: {question}"
    )
    json_parser = SimpleJsonOutputParser()
    json_chain = json_prompt | llm | json_parser
    response = json_chain.invoke({"question": template})
    report_title = response['title']
#    chain = prompt | llm
#    log.debugLog('Calling getReportTitle', state=state)
#    response = chain.invoke(chathistory)
#    log.debugLog(response, state=state)
#    report_title = response.content.split('=')[1]
    state['process']['steps'].append(log.llmCallLog(llm          = llm,
                                                         prompt       = template,
                                                         input        = chathistory,
                                                         output       = response,
                                                         parsedOutput = {
                                                             'report title' : report_title,
                                                         },
                                                         purpose      = 'generate a title for a pdf (note: using pydantic parser)'
                                                        )
                                         )
    log.debugLog(report_title, state=state)
    return report_title

def getReportSummary(chathistory, state):
    """
    Uses a language model to generate a summary section for a report based on the chat session with BRAD.

    Args:
        chathistory (dict): A dictionary containing the chat history with user inputs and model outputs.
        state (dict): A dictionary containing the current chat status, including the language model instance 
                           and user prompt.

    Returns:
        str: The generated summary section for the report based on the chatlog and user prompt.
    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: July 2, 2024
    llm      = state['llm']
    template = """You are in charge of writing a report to summarize a chatbot session that performed a series of steps in a bioinformatics workflow. Given the following chatlog, write a section of to summarize this. Your summary section should be at least three sentences long (minimum) and no longer than twelve sentences.
    
This report is in reposnse to the following user request:
{userinput}

**Chatlog Overview:**
{{chatlog}}

Format your output exactly as follows. Do not include additional text/characters.
**Output:**
Summary=<put the title title here>
"""
    template = template.format(userinput=state['prompt'])
    print(template)
    print(f"{chathistory=}")
    prompt = PromptTemplate(template=template, input_variables=["chatlog"])
    chain = prompt | llm
    log.debugLog('Calling getReportSummary', state=state)
    response = chain.invoke(chathistory)
    log.debugLog(response, state=state)
    report_summary = response.content.split('=')[1]
    state['process']['steps'].append(log.llmCallLog(llm          = llm,
                                                         prompt       = prompt,
                                                         input        = chathistory,
                                                         output       = response,
                                                         parsedOutput = {
                                                             'report summary' : report_summary,
                                                         },
                                                         purpose      = 'generate a summayr of the chat history'
                                                        )
                                         )
    log.debugLog(report_summary, state=state)
    return report_summary


def getChatInputOutputs(state, chatlog):
    """
    Creates a single string to summarize the chat history between a user and BRAD.

    Args:
        state (dict): A dictionary containing the current chat status, not directly used in this function.
        chatlog (dict): A dictionary representing the chat history with keys as indices and values as dictionaries
                        containing 'prompt' and 'output' keys.

    Returns:
        str: A formatted string summarizing the chat history with inputs from the user and outputs from BRAD.
    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: July 2, 2024
    chathistory = "**Chat History**\n\n"
    for i in chatlog.keys():
        chathistory += "Human: " + chatlog[i]['prompt'] + '\n\n'
        chathistory += "BRAD: " + chatlog[i]['output'] + '\n\n\n'
    return chathistory



def getLatexTemplate(path2latexTemplate):
    """
    Reads the LaTeX template file and returns it as a string.
    
    Args:
    path2latexTemplate (str): The path to the LaTeX template file.
    
    Returns:
    str: The contents of the LaTeX template file as a string.
    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: July 2, 2024
    try:
        with open(path2latexTemplate, 'r') as file:
            latex_template = file.read()
        return latex_template
    except FileNotFoundError:
        return f"Error: The file at {path2latexTemplate} was not found."
    except IOError:
        return f"Error: An error occurred while reading the file at {path2latexTemplate}."


def summarizeSteps(state):
    """
    .. warning: we will remove this soon
    
    This function should write a report about what happened in a pipeline
    """
    query    = state['prompt']
    llm      = state['llm']              # get the llm

    chatlog = json.load(open(os.path.join(state['output-directory'], 'log.json')))

    # Write the report in this outline
    #  - Title
    title = setTitle(state, chatlog)
    
    #  - Driving prompt
    prompt = getPrompt(state, chatlog)

    #  - Summary of process
    processSummary = getProcessSummary(state, chatlog)
    
    #  - Summary of outputs
    reportBody = getReportBody(state, chatlog)
    
    #  - References
    references = getReferences(state, chatlog)

    # Fill in the report
    report = getFirstLatexReportOutline(title=title)

    report = report.format(prompt=prompt,
                           processSummary=processSummary,
                           reportBody=reportBody,
                           references=references)

    # Report to pdf
    state, report_file = reportToPdf(state, report)
    state = utils.compile_latex_to_pdf(state, report_file)
    
    return state #, report

def reportToPdf(state, report):
    """
    Writes the formatted report to a PDF file.

    Args:
        state (dict): A dictionary containing the current chat status, used to track process steps and file paths.
        report (str): The formatted report content in LaTeX format.

    Returns:
        tuple: A tuple containing the updated chat status dictionary and the filename of the generated PDF report.
    """
    report = ensureLatexFormatting(report)
    report_file = 'REPORT.tex'
    state = utils.save(state, report, report_file)
    report_file = state['process']['steps'][-1]['new file']
    return state, report_file

def ensureLatexFormatting(report):
    """
    .. warning: we need to make this function work better
    
    This function ensures appropriate LaTeX formatting.
    
    For instance:
        - any "_" not in a math environment should be immediately proceeded by "\" to make them all "\_". 
          If they are already immediately proceeded by "\", then no change is required.
        - all math should be in an appropriate math environment including $ for intext characters or 
          \begin{equation} <math here> \end{equation} for full lines of math.
        - any file names that have .csv, .tsv, .pkl, .h5ad, .txt, .py, .cpp should be in the \texttt{<file name>} environment.
    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: July 2, 2024

    # Function to escape underscores not already escaped
    def escape_underscores(text):
        return re.sub(r'(?<!\\)_', r'\\_', text)

    # Process the report to handle different parts
    parts = re.split(r'(\$.*?\$)', report, flags=re.DOTALL)
    
    processed_parts = []
    for part in parts:
        if part.startswith('$') and part.endswith('$'):
            # Inline math, do not change underscores
            processed_parts.append(part)
        else:
            # Regular text, escape underscores and wrap math expressions in $
            part = escape_underscores(part)
            part = re.sub(r'(\b[a-zA-Z]\w*\s*=\s*[^.]+\b)', r'$\1$', part)  # Heuristic to detect simple equations
            processed_parts.append(part)
    
    report = ''.join(processed_parts)

    # Ensure file names with specific extensions are in \texttt{}
    report = re.sub(r'(\S+\.(csv|tsv|pkl|h5ad|txt|py|cpp))', r'\\texttt{\1}', report)
    
    return report

    

def setTitle(state, chatlog):
    """
    .. warning: we will remove this soon
    
    This function sets the title of the report based on the initial user query
    """

    # Get the pro
    for promptNumber in chatlog.keys():
        log.debugLog(promptNumber, state=state) 
        if promptNumber == 'llm':
            continue
        if chatlog[promptNumber]['process']['module'] == 'PLANNER':
            prompt = chatlog[promptNumber]['prompt']
            break

    # Set the title
    template = setReportTitleTemplate()
    PROMPT = PromptTemplate(input_variables=["user_query"], template=template)
    chain = PROMPT | state['llm']
    res = chain.invoke(prompt)
    log.debugLog('Title Setter Output: \n\n' + str(res.content), state)
    title   = res.content.split('=')[1].strip()
    log.debugLog('Title=' + str(title), state)
    return title

def getPrompt(state, chatlog):
    """
    .. warning: we will remove this soon
    
    This function finds the original prompt used to generate this report
    """
    for promptNumber in chatlog.keys():
        if promptNumber == 'llm':
            continue
        if chatlog[promptNumber]['process']['module'] == 'PLANNER':
            prompt = chatlog[promptNumber]['prompt']
            return prompt
    return ""

def getProcessSummary(state, chatlog):
    """Summarizes each stage of the process"""
    for promptNumber in chatlog.keys():
        if promptNumber == 'llm':
            continue
        if chatlog[promptNumber]['process']['module'] == 'PLANNER':
            fullPlanner = str(chatlog[promptNumber]['status']['queue'])
    log.debugLog(fullPlanner, state=state) 
    template = summarizeAnalysisPipelineTemplate()
    PROMPT = PromptTemplate(input_variables=["pipeline"], template=template)
    chain = PROMPT | state['llm']
    res = chain.invoke(fullPlanner)
    log.debugLog('Process Summary Output: \n\n' + str(res.content), state)
    processSummary   = res.content.split('=')[1].strip()
    log.debugLog('Process Summary=' + str(processSummary), state)
    return processSummary

def getReportBody(state, chatlog):
    """Summarizes individual steps"""
    reportBody = ""
    for promptNumber in chatlog.keys():
        if promptNumber == 'llm':
            continue
        # Add to the report body based on the module
        module = chatlog[promptNumber]['process']['module']
        if module == 'RAG':
            reportBody += ragReporter(chatlog[promptNumber], state)
        elif module == 'CODE':
            reportBody += codeReporter(chatlog[promptNumber], state=state)
        elif module == 'DATABASE':
            reportBody += databaseReporter(chatlog[promptNumber], state)
        # Add to the report body based on the elements
        # reportBody += addFigures(chatlog[promptNumber])
        # reportBody += databaseReporter(chatlog[promptNumber])
    return reportBody

def codeReporter(chatlogStep, currentreport='', state=None):
    """Summarizes running of some code"""
    scriptRun = chatlogStep['process']['steps'][0]['parsedOutput']['scriptName']
    docstring = read_python_docstrings(scriptRun)
    code      = chatlogStep['process']['steps'][1]['parsedOutput']['code']
    output    = chatlogStep['output']

    template = """You are responsible for compiling a report of a bioinformatics pipeline. In the current section you are writing, you must summarize the output of running a piece of code in the pipeline. You will be provided with the current report, the doc strings of the function that was run, the command used to run the script, and the output of the script. Then, please provide the next step of paragraphs(s), in appropriate latex format, to summarize this stage of the pipeline.

**Current Report**
{currentreport}

**Documentation of the function that was run**
{docstring}

**Executed Code**
{code}

**Code Output**
{codeoutput}

**Aditional Instructions**
{{userprompt}}

Please format your output exactly as follows.
**Output**
Summary: <put summary here>
"""
    template = template.format(currentreport=currentreport,
                               docstring=docstring,
                               code=code,
                               codeoutput=output,
    )
    print(template)
    PROMPT = PromptTemplate(input_variables=["userprompt"], template=template)
    chain = PROMPT | state['llm']
    response = chain.invoke(chatlogStep['prompt'])
    print(response)
    return response.content # response.content.split('=')[1]

def ragReporter(chatlogStep, state):
    """Summarize the output of the RAG stage"""
    output = chatlogStep['output']
    template = summarizeRAGTemplate()
    log.debugLog(template, display=True)
    # template = template.format(output=output)
    PROMPT = PromptTemplate(input_variables=["output"], template=template)
    chain = PROMPT | state['llm']
    res = chain.invoke(output)
    processSummary   = res.content.split('=')[1].strip().replace('_', '\_')
    return processSummary
    
def databaseReporter(chatlogStep, state):
    """Summarize the output of the DATABASE lookup stage"""
    databaseTex = ""
    prompt = chatlogStep['prompt']
    output = chatlogStep['output']

    for step in chatlogStep['process']['steps']:
        # Check if the step saved a figure
        if 'func' in step.keys() and step['func'] == 'utils.save':
            filename = step['new file']
            state = log.userOutput(filename, state=state) # 
            df = utils.load_file_to_dataframe(filename)
    if df is not None:
        numrows = df.shape[0]
    else:
        numrows = 0
    
    template = summarizeDatabaseCallerTemplate()
    log.debugLog(template, display=True)
    template = template.format(output=output, numrows=numrows)
    PROMPT = PromptTemplate(input_variables=["input"], template=template)
    chain = PROMPT | state['llm']
    res = chain.invoke(prompt)
    processSummary   = res.content.split('=')[1].strip().replace('_', '\_')
    databaseTex += processSummary
    databaseTex += addFigures(chatlogStep, state)
    databaseTex += addTables(chatlogStep, state)
    
    return databaseTex

def addFigures(chatlogStep, state):
    """Add figures to the latex output file"""
    figureTex = ""
    for step in chatlogStep['process']['steps']:
        # Check if the step saved a figure
        if 'func' in step.keys() and step['func'] == 'utils.savefig':
            figureCode = """\n\n\\begin{figure}
    \\centering
    \\includegraphics[width=\\textwidth]{""" + step['new file'] + """}
\\end{figure}\n\n"""
            figureTex += figureCode
    return figureTex
# \\caption{\\textbf{""" +step['new file'] + """}}
# \\label{fig:enter-label}

def addTables(chatlogStep, state):
    """Add tables to the latex output file"""
    tableTex = ""
    df = None
    # Choose table
    for step in chatlogStep['process']['steps']:
        # Check if the step saved a figure
        if 'func' in step.keys() and step['func'] == 'utils.save':
            filename = step['new file']
            state = log.userOutput(filename, state=state) 
            df = utils.load_file_to_dataframe(filename)
            break
    if df is None:
        return ""
    log.debugLog(type(df), state=state) 
    log.debugLog(df.head(3), state=state) 
    # df = pd.read_csv(filename)
    # Build latex
    tableTex += ('\n\n' + dataframe_to_latex(df) + '\n\n')
    return tableTex

def getReferences(state, chatlog):
    """Finds all documents used in the chat"""
    referenceList = []
    for promptNumber in chatlog.keys():
        if promptNumber == 'llm':
            continue
        if chatlog[promptNumber]['process']['module'] == 'RAG':
            sources = chatlog[promptNumber]['process']['sources']
            for source in sources:
                referenceList.append(source)
    references = '\\begin{itemize}\n'
    for ref in referenceList:
        references += '\\item ' + ref + '\n'
    references += '\\end{itemize}'
    return references

def dataframe_to_latex(df):
    """Converts a dataframe to latex"""
    df = df.head(20)
    for col in df.columns:
        if len(df[col].unique()) == 1:
            df.drop(col,inplace=True,axis=1)
    max_length = 250
    
    # Identify columns to drop
    columns_to_drop = [col for col in df.columns if df[col].astype(str).apply(len).max() > max_length]
    
    # Drop identified columns
    df = df.drop(columns=columns_to_drop)
    
    df['path_name'] = df['path_name'].apply(lambda x: x[:35] + "...")

    # Generate LaTeX table
    latex_table = tabulate(df, tablefmt='latex', showindex=False, headers='keys')


    # Add booktabs and other LaTeX formatting
    latex_table = latex_table.replace("\\\\ ", "\\").replace("toprule", "toprule\nrank & path-name & p-val & z-score & combined-score & adj-p-val \\\\ \\midrule").replace("midrule", "midrule\n").replace("bottomrule", "bottomrule\n")

    # Add table environment
    latex_table = "\\begin{table}[h]\n\\centering\n" + latex_table + "\n\\end{table}"

    return latex_table

def getFirstLatexReportOutline(title='BRAD Output'):
    """A hardcoded latex template for generating a pdf report"""
    report = """\\documentclass{{article}}
\\usepackage[margin=1in]{{geometry}}
\\usepackage{{graphicx}}
\\usepackage{{url}}
\\begin{{document}}

\\title{{""" + title + """}}
\\author{{BRAD}}
\\maketitle

\\section*{{User Prompt}}
{prompt}

\\section*{{Summary of Process}}
{processSummary}

\\section*{{Summary of Outputs}}
{reportBody}

\\section*{{References}}
{references}

\\end{{document}}
"""
    return report


"""
The quickest route to turning on BRAD is to open a command line interface (CLI) chat session. This creates a
chatbot instance and lets the user message back and forth from the terminal.

Running BRAD from the CLI
-------------------------

From the root of the project, the code required to open this chat session is minimal:

>>> python BRAD/chat.py // Minimal required command

This command creates an interactive `Agent` and puts the `Agent` in the `chat` mode.
Running this command will use all of the default settings to open a chat session
from the command line. This will prompt you to enter an OpenAI key, and display the
following messages:

>>> Enter your Open AI API key: █████
... Would you like to use a database with BRAD [Y/N]?
... █████
... Welcome to RAG! The chat log from this conversation will be saved to /home/jpic/BRAD/2024-10-08_15-48-38/log.json. How can I help?
... ==================================================
>>> Input >> █████

In the sections that follow, we will show:

1. How the `Agent` class is organized and works
2. The GUI and other programatic interfaces to use the BRAD system

Available Methods
-----------------

"""

import os
import sys

# Enusre the BRAD package is accessible, even if not installed fully
parent_directory = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.append(parent_directory)

from BRAD.agent import *

def chat(
        model_path = '/nfs/turbo/umms-indikar/shared/projects/RAG/models/llama-2-7b-chat.Q8_0.gguf',
        persist_directory = "/nfs/turbo/umms-indikar/shared/projects/RAG/databases/DigitalLibrary-10-June-2024/",
        llm=None,
        ragvectordb=None,
        embeddings_model=None,
        restart=None,
        name='BRAD',
        max_api_calls=None,
        config=None
    ):
    """
    To interact with a BRAD Agent, this method instantiates a new agent, with all of the specified functionality, and uses the `chat()` method.
    This allows a user to interface with the code without creating an agent.

    :param model_path: The path to the Llama model file, defaults to '/nfs/turbo/umms-indikar/shared/projects/RAG/models/llama-2-7b-chat.Q8_0.gguf'.
    :type model_path: str, optional
    :param persist_directory: The directory where the literature database is stored, defaults to "/nfs/turbo/umms-indikar/shared/projects/RAG/databases/Transcription-Factors-5-10-2024/".
    :type persist_directory: str, optional
    :param llm: The language model to be used. If None, it will be loaded within the function.
    :type llm: PreTrainedModel, optional
    :param ragvectordb: The RAG vector database to be used. If None, it will prompt the user to load it.
    :type ragvectordb: Chroma, optional
    :param embeddings_model: The embeddings model to be used. If None, it will be loaded within the function.
    :type embeddings_model: HuggingFaceEmbeddings, optional

    :raises FileNotFoundError: If the specified model or database directories do not exist.
    :raises json.JSONDecodeError: If the configuration file contains invalid JSON.
    :raises KeyError: If required keys are missing from the configuration or chat status.

    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: June 4, 2024

    # Dev. Comments:
    # -------------------
    # This function executes a chat session
    #
    # History:
    # - 2024-06-04: wrote 1st draft of this code to execute a series of prompts
    #               or a conversation between the user and BRAD
    #
    #       ...     modified this funciton adding new features, logging, routing
    #               and more
    #
    # - 2024-07-10: refactored brad.py file to a class and split this method
    #               into the chatbot.__init__(), chatbot.invoke(), and
    #               chatbot.chat() methods. This method was maintained so that
    #               a user can still fun `brad.chat()` easily and without knowledge
    #               of the class structure

    bot = Agent(
        model_path = model_path,
        persist_directory = persist_directory,
        llm=llm,
        ragvectordb=ragvectordb,
        embeddings_model=embeddings_model,
        restart=restart,
        name=name,
        max_api_calls=max_api_calls,
        config=config
    )
    bot.chat()


if __name__ == "__main__":
    chat()




"""
This module provides functions for generating and managing agentic workflows or pipelines with multiple steps
each to be executed by individual tool modules. The main method, `planner`, builds pipelines based on user input
and the available modules. It also supports selecting pre-existing pipelines or designing new ones 
when necessary.
"""


import os
import re
import json
import difflib

from langchain.memory import ConversationBufferMemory
from langchain.chains import ConversationChain, LLMChain
from langchain.prompts import PromptTemplate
from langchain_core.prompts.prompt import PromptTemplate

from BRAD.promptTemplates import plannerTemplate, plannerEditingTemplate, plannerTemplateForLibrarySelection
from BRAD import log

def planner(state):
    """
    Generates a plan based on the user prompt using a language model, allows the user 
    to review and edit the plan, and then updates the state with the finalized plan.

    Args:
        state (dict): A dictionary containing the LLM, user prompt, vector database, 
                           memory, and configuration settings for the planning process.

    Returns:
        dict: The updated state containing the finalized plan and any modifications 
              made during the process.

    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: June 16, 2024

    # Dev. Comments:
    # -------------------
    # This function initializes a chat session and the state variable
    #
    # History:
    # - 2024-06-16: 1st draft of this method
    # - 2024-07-25: this function is refactors to allow the planner to save
    #               new pipelines and rerun old pipelines.
    #
    # Issues:
    # - The parsing of new/custom pipelines into prompts and a queue doesn't work
    #   all that well
    # - The queue should be implemented with a class rather than with only a list,
    #   IP, and set of prompts
    #
    # TODOs:
    # - add code to let BRAD fill in the template of a prebuilt pipeline
    # - add code to save new pipelines

    llm      = state['llm']              # get the llm
    prompt   = state['prompt']           # get the user prompt
    memory   = state['memory']           # get the memory of the model

    # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # Select to use a prebuilt tempalte or design out own
    # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    template = plannerTemplateForLibrarySelection()
    pipelines, pipelineSummary = getKnownPipelines(state)
    template = template.format(pipeline_list=pipelineSummary)
    PROMPT = PromptTemplate(input_variables=["input"], template=template)
    conversation = LLMChain(prompt  = PROMPT,
                            llm     = llm,
                            verbose = state['config']['debug'],
                            )
    response = conversation.predict(input=prompt)
    log.debugLog(response, state=state)
    pipelineSelection = response.split('\n')[0].split(':')[1]
    pipeline_names = [name.upper() for name in pipelines.keys()]
    pipeline_names.append('CUSTOM')
    selected_pipeline = difflib.get_close_matches(pipelineSelection.upper(), pipeline_names, n=1, cutoff=0.0)
    if len(selected_pipeline) == 0:
        selected_pipeline = "CUSTOM"
    else:
        selected_pipeline = selected_pipeline[0]
    log.debugLog(f'selected_pipeline={selected_pipeline}', state=state)

    state['process']['steps'].append(
        log.llmCallLog(
            llm          = llm,
            prompt       = PROMPT,
            input        = prompt,
            output       = response,
            parsedOutput = {
                'selected pipeline': selected_pipeline
            },
            purpose      = 'determine if a known pipeline can be used or a new one si required'
        )
    )

    # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # Building a custom pipeline
    # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    if selected_pipeline == 'CUSTOM':
        template = plannerTemplate()
        PROMPT = PromptTemplate(input_variables=["history", "input"], template=template)
        conversation = ConversationChain(prompt  = PROMPT,
                                         llm     = llm,
                                         verbose = state['config']['debug'],
                                         memory  = memory,
                                        )
        response = conversation.predict(input=prompt)
        response += '\n\n'
        state = log.userOutput(response, state=state)
        state['process']['steps'].append(log.llmCallLog(llm          = llm,
                                                             prompt       = PROMPT,
                                                             memory       = memory,
                                                             input        = prompt,
                                                             output       = response,
                                                             parsedOutput = {
                                                                 'output' : response,
                                                             },
                                                             purpose      = 'generate a proposed set of prompts'
                                                            )
                                             )
        while True:
            state = log.userOutput('Do you want to proceed with this plan? [Y/N/edit]', state=state)
            prompt2 = input('Input >> ')
            state['process']['steps'].append(
                {
                    'func'           : 'planner.planner',
                    'prompt to user' : 'Do you want to proceed with this plan? [Y/N/edit]',
                    'input'          : prompt2,
                    'purpose'        : 'get new user input'
                }
            )
            if prompt2 == 'Y':
                break
            elif prompt2 == 'N':
                return state
            else:
                template = plannerEditingTemplate()
                template = template.format(plan=response)
                log.debugLog(template, state=state)
                PROMPT   = PromptTemplate(input_variables=["user_query"], template=template)
                chain    = PROMPT | llm
                
                # Call chain
                response = chain.invoke(prompt2).content.strip() + '\n\n'
                state['process']['steps'].append(log.llmCallLog(llm          = llm,
                                                                     prompt       = PROMPT,
                                                                     # memory should be included in this chain
                                                                     # memory       = memory,
                                                                     input        = prompt2,
                                                                     output       = response,
                                                                     parsedOutput = {
                                                                         'output' : response,
                                                                     },
                                                                     purpose      = 'update the proposed set of prompts'
                                                                    )
                                                     )
                state = log.userOutput(response, state=state)

        processes = response2processes(response)
        log.debugLog(processes, state=state)
        state['process']['steps'].append(
            {
                'func' : 'planner.planner',
                'what' : 'designed a new pipeline'
            }
        )

        # Check if the chat is interactive
        if state['interactive']:
            # Prompt the user to decide if they want to save the pipeline to a file
            state = log.userOutput('Would you like to save this pipeline to a file? [Y/N]', state=state)
            saveNewPipeline = input(">>>")
        
            # If the user chooses to save the pipeline
            if saveNewPipeline.upper() == "Y":
                # Prompt the user to enter a name for the pipeline
                state = log.userOutput('Enter a name for your pipeline', state=state)
                fname = input(">>>")
        
                # Prompt the user to enter a description for the pipeline
                state = log.userOutput('Enter a description of your pipeline', state=state)
                description = input(">>>")
        
                # Create a dictionary to store the pipeline information
                pipelineJSONdict = {
                    'name': fname,  # Use 'fname' instead of 'name' to match the input
                    'description': description,
                    'queue': processes  # Fixed typo: 'proecesses' to 'processes'
                }
                for k in range(len(pipelineJSONdict['queue'])):
                    pipelineJSONdict['queue'][k]['output'] = []
        
                # Get the directory path to save the pipeline file
                pipelines_dir = state['config']['PLANNER']['path']
        
                # Construct the full file path
                filepath = os.path.join(pipelines_dir, f"{fname}.json")
        
                # Save the pipelineJSONdict to a file
                try:
                    with open(filepath, 'w') as f:
                        json.dump(pipelineJSONdict, f, indent=4)
                    state = log.userOutput(f"Pipeline saved successfully to {filepath}", state=state)
                except Exception as e:
                    state = log.userOutput(f"Error saving pipeline: {e}", state=state)

    # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # Parameterize a predesigned pipeline
    # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    else:
        pipeline = pipelines[selected_pipeline]
        # Initialize processes dictionary from the loaded pipeline queue
        loadedProcesses, processes = pipeline['queue'], {}
        
        # If loadedProcesses is a dictionary
        if isinstance(loadedProcesses, dict):
            # Iterate over the items in the dictionary
            for key, value in loadedProcesses.items():
                # Convert the key to an integer if it is a string representation of a digit
                new_key = int(key) if isinstance(key, str) and key.isdigit() else key
                processes[new_key] = value
        
        # If loadedProcesses is a list
        elif isinstance(loadedProcesses, list):
            # Iterate over the list with indices
            for key, value in enumerate(loadedProcesses):
                processes[key] = value
                
        # If loadedProcesses is neither a dictionary nor a list, issue a warning
        else:
            log.debugLog("Warning: loadedProcesses is neither a dictionary nor a list", state=state)
        print('Display pipeline')
        state = displayPipeline2User(processes, state=state)
        state['process']['steps'].append(
            {
                'func' : 'planner.planner',
                'what' : 'loaded an older pipeline'
            }
        )

    state['interactive'] = False
    state['queue'] = processes
    state['queue pointer'] = 1 # the 0 object is a place holder
    state['process']['steps'].append(
        {
            'func' : 'planner.planner',
            'what' : 'set the queue and set the queue pointer to 1'
        }
    )
    return state

def displayPipeline2User(process, state=None):
    """
    Displays the steps of the process pipeline to the user, logging each step.
    
    This function iterates through the steps of a process pipeline, outputs each step to the user in a
    standardized format, and updates the chat status with the logged outputs. Each step is labeled as
    **Step X**, where X is the key, followed by the corresponding value of the process step.
    
    :param process: A dictionary representing the process pipeline. Each key-value pair corresponds 
                    to a step in the process, where the key is the step number or name, and the value 
                    is the step's description or details.
    :type process: dict
    :param state: The current chat status dictionary to which the output will be appended. If not 
                       provided, a default value of `None` is used.
    :type state: dict, optional
    
    :return: The updated chat status after logging all process steps.
    :rtype: dict
    
    """
    for key, value in process.items():
        state=log.userOutput("** Step " + str(key) + "**", state=state)
        state=log.userOutput(str(value), state=state)
        state=log.userOutput("\n", state=state)
    return state

def response2processes(response):
    """
    Converts a response string into a list of processes, each with an order, module, 
    prompt, and description. It identifies modules from a predefined list and parses 
    the response into steps based on these modules.
    
    Args:
        response (str): The response string containing the steps and corresponding details.
    
    Returns:
        list: A list of dictionaries, each representing a process with the following keys:
              - 'order': The order of the process.
              - 'module': The module associated with the process.
              - 'prompt': The prompt to invoke the process.
              - 'description': A description of the process.
    
    Example:
        >>> response = '''
        ... **Step 1: RAG**
        ... Prompt: Retrieve documents related to AI research
        ... **Step 2: SCRAPE**
        ... Prompt: Scrape data from the specified website
        ... '''
        >>> processes = response2processes(response)
        >>> print(processes)
        [
            {
                'order': 0,
                'module': 'PLANNER',
                'prompt': None,
                'description': 'This step designed the plan. It is placed in the queue because we needed a placeholder for 0-indexed lists.',
            },
            {
                'order': 1,
                'module': 'RAG',
                'prompt': '/force RAG Retrieve documents related to AI research',
                'description': '**Step 1: RAG\\nPrompt: Retrieve documents related to AI research\\n',
            },
            {
                'order': 2,
                'module': 'SCRAPE',
                'prompt': '/force SCRAPE Scrape data from the specified website',
                'description': '**Step 2: SCRAPE\\nPrompt: Scrape data from the specified website\\n',
            }
        ]
    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: June 16, 2024

    # History:
    # - 2024-06-16: 1st draft of this method
    # - 2024-08-02: this was changed to split the prompts at the word "Step" as
    #               opposed to "**Step ". It remains a problem that this will be
    #               very brittle to the use by different LLMs.
    #
    # Issues:
    # - This is VERY brittle to use by different LLMs
    modules = ['RAG', 'SCRAPE', 'DATABASE', 'CODE', 'WRITE', 'ROUTER']
    stageStrings = response.split('Step')
    processes = [
        {
            'order'  : 0,
            'module' : 'PLANNER',
            'prompt' : None,
            'description' : 'This step designed the plan. It is placed in the queue because we needed a place holder for 0 indexed lists.',
        }
    ]
    print(stageStrings)
    for i, stage in enumerate(stageStrings):
        stageNum = i
        found_modules = [module for module in modules if module in stage]
        if len(found_modules) == 0:
            continue
        prompt = re.findall(r'Prompt: (.*?)\n', stage)
        for module in found_modules:
            
            processes.append({
                'order':stageNum,
                'module':module,
                'prompt':'/force ' + module + ' ' + stage, # + prompt[0],
                'description':stage,
            })
    return processes

def getKnownPipelines(state):
    """
    This function reads all available pipeline JSON files in the 'pipelines' directory
    and extracts their 'name' and 'description' fields. It formulates a summary string
    that can be used as input to an LLM prompt for selecting the appropriate pipeline.

    Returns:
        tuple: A tuple containing two elements:
            - pipelines (list): A list of dictionaries where each dictionary represents
              a pipeline read from a JSON file.
            - summary (str): A formatted string summarizing the 'name' and 'description'
              of each pipeline.
    
    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: July 25, 2024
    
    # Get the path to the 'pipelines' directory
    # current_script_path = os.path.abspath(__file__)
    # current_script_dir = os.path.dirname(current_script_path)
    # pipelines_dir = os.path.join(current_script_dir, 'pipelines')
    pipelines_dir = state['config']['PLANNER']['path']

    # Initialize an empty list to store pipeline dictionaries
    pipelines = {}
    summary = ""
    # Read all JSON files in the 'pipelines' directory
    for file_name in os.listdir(pipelines_dir):
        if file_name.endswith('.json'):
            file_path = os.path.join(pipelines_dir, file_name)
            with open(file_path, 'r') as file:
                pipeline = json.load(file)
                pipelines[pipeline['name'].upper()] = {
                    'path' : file_path,
                    'description': pipeline['description'],
                    'queue': pipeline['queue']
                }
                # Extract 'name' and 'description' to build the summary string
                summary += f"Name: {pipeline['name']}\tDescription: {pipeline['description']}\n"
    return pipelines, summary



"""

Gene Ontology (GO)
------------------

This module provides functions to perform `Gene Ontology (GO) <https://geneontology.org/>`_ searches, download charts, and retrieve associated publications and annotations based on gene terms. The module interacts with external APIs, such as QuickGO and PubMed, to gather the relevant information.

Available Methods
~~~~~~~~~~~~~~~~~

This module has the following methods:

"""

import requests, sys
import pandas as pd
import json
import csv
from requests_html import HTMLSession
import requests
from requests.exceptions import ConnectionError
import os
import copy
from io import StringIO

from BRAD import log

def geneOntology(state, goQuery):
    """
    Performs Gene Ontology (GO) search for specified genes and updates the chat status with the results.

    :param goQuery: The query string containing gene names or terms for GO search.
    :type goQuery: str
    
    :param state: The current status of the chat, including the prompt, configuration, and process details.
    :type state: dict

    :raises FileNotFoundError: If the gene list file is not found.

    :return: The updated chat status dictionary containing the GO search results and process details.
    :rtype: dict

    """
    # Auth: Marc Choi
    #       machoi@umich.edu
    current_script_path = os.path.abspath(__file__)
    current_script_dir = os.path.dirname(current_script_path)
    file_path = os.path.join(current_script_dir, 'helperData', 'gene_list.txt')
    with open(file_path, 'r') as file:
        contents = file.read()
    gene_list = contents.split('\n')
    if len(goQuery) > 0:
        go_process = goSearch(goQuery, state)
        state['process']['GO'] = go_process
    return state
            
def goSearch(query, state):
    """
    Performs a search on Gene Ontology (GO) based on the provided query and allows downloading associated charts and papers.

    :param query: The query list containing gene names or terms for GO search.
    :type query: list

    :return: A dictionary containing the GO search process details.
    :rtype: dict

    """
    # Auth: Marc Choi
    #       machoi@umich.edu
    process = {}
    output = {}
    for terms in query:
        output, geneStatus = textGO(terms, state)
        process['output'] = output
        if geneStatus == True:
            state = log.userOutput('\n would you like to download charts associated with these genes [Y/N]?', state=state)
            for term in output:
                go_id = str(term[0])
                chartGO(go_id, state)
                state = log.userOutput('\n would you like to download the paper associated with these genes [Y/N]?', state=state)
                # download2 = input().strip().upper()
                # process['paper_download'] = (download2 == 'Y')
                # if download2 == 'Y':
                pubmedPaper(go_id, state)

                    
        else:
            state = log.userOutput('\n would you like to download the gene product annotation [Y/N]?', state=state)
            for term in query:
                state = log.userOutput(term, state=state)
                annotations(term, state)
    return process


def textGO(query, state):
    """
    Performs a text-based Gene Ontology (GO) search for a specified query and returns the extracted data and gene status.

    :param query: The query string containing gene names or terms for GO search.
    :type query: str

    :raises requests.HTTPError: If the HTTP request to the GO API fails.

    :return: A tuple containing the extracted data and a boolean indicating whether the query is a gene.
    :rtype: tuple

    """
    # Auth: Marc Choi
    #       machoi@umich.edu
    gene = True
    requestURL = "https://www.ebi.ac.uk/QuickGO/services/ontology/go/search?query="+query+"&limit=25&page=1"
    r = requests.get(requestURL, headers={ "Accept" : "application/json"})

    if not r.ok:
        r.raise_for_status()
        sys.exit()
    extracted_data = []

    responseBody = r.text
    data = json.loads(responseBody)
    if data['numberOfHits'] == 0:
        gene=False
        state = log.userOutput("No Gene Ontology Available - Searching Gene Products", state=state)
        requestURL = requestURL = "https://www.ebi.ac.uk/QuickGO/services/geneproduct/search?query="+query
        r = requests.get(requestURL, headers={ "Accept" : "application/json"})

        if not r.ok:
            r.raise_for_status()
            sys.exit()

        responseBody = r.text
        data = json.loads(responseBody)
        for result in data['results']:
            id = result['id']
            extracted_data.append(id)
        for id in extracted_data:
            state = log.userOutput(f"ID: {id}", state=state)
    else:
        for result in data['results']:
            id = result['id']
            text = result['definition']['text']
            extracted_data.append((id, text))
            # Print the extracted data
        for id, text in extracted_data:
            state = log.userOutput(f"ID: {id}", state=state)
            state = log.userOutput(f"Text: {text}", state=state)
    return extracted_data, gene

    

#Input is a GO:----- identification for a gene        
def chartGO(identifier, state):

    """
    Downloads a chart for a specified Gene Ontology (GO) identifier.

    :param identifier: The GO identifier for which the chart is to be downloaded.
    :type identifier: str

    :raises requests.HTTPError: If the HTTP request to download the chart fails.

    """
    # Auth: Marc Choi
    #       machoi@umich.edu
    try: 
        path = os.path.abspath(os.getcwd()) + '/go_charts'
        os.makedirs(path, exist_ok = True)
        state = log.userOutput("Directory '%s' created successfully" % path, state=state)
    except OSError as error:
        state = log.userOutput("Directory '%s' can not be created" % path, state=state)
    requestURL = "https://www.ebi.ac.uk/QuickGO/services/ontology/go/terms/{ids}/chart?ids=GO%3A"+identifier[3:]
    img_data = requests.get(requestURL).content

    # save chart
    with open(os.path.join(path, identifier[3:] + '.jpg'), 'wb') as handler:
        handler.write(img_data)

#Input is a GO:----- identification for a gene     
def pubmedPaper(identifier, state):
    """
    Downloads PubMed papers associated with a specified Gene Ontology (GO) identifier.

    :param identifier: The GO identifier for which the associated PubMed papers are to be downloaded.
    :type identifier: str

    :raises requests.HTTPError: If the HTTP request to the PubMed API fails.

    """
    # Auth: Marc Choi
    #       machoi@umich.edu
    
    requestURL2 = "https://www.ebi.ac.uk/QuickGO/services/ontology/go/terms/GO%3A"+identifier[3:]
    r = requests.get(requestURL2, headers={ "Accept" : "application/json"})
    if not r.ok:
        r.raise_for_status()
        sys.exit()
    responseBody = r.text
    data = json.loads(responseBody)
    # Extract the dbId from the first result
    dbId = []
    
    #state = log.userOutput(data, state=state)
    if data['numberOfHits'] > 0:
        xrefs = data['results'][0].get('definition', {}).get('xrefs', [])
        if xrefs:
            for db in xrefs:
                dbId.append(db['dbId'])
        s = HTMLSession()
        headers = {'user-agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36'}
        if len(dbId) > 0:
            try: 
                path = os.path.abspath(os.getcwd()) + '/specialized_docs'
                os.makedirs(path, exist_ok = True) 
                state = log.userOutput("Directory '%s' created successfully" % path, state=state)
            except OSError as error: 
                state = log.userOutput("Directory '%s' can not be created" % path, state=state)
            for idname in dbId:
                state = log.userOutput(idname, state=state)
                try:
                    base_url = 'https://pubmed.ncbi.nlm.nih.gov/'
                    r = s.get(base_url + idname + '/', headers = headers, timeout = 5)
                    try:
                        pdf_url = r.html.find('a.id-link', first=True).attrs['href']
                        state = log.userOutput(pdf_url, state=state)
                        if "doi" in pdf_url:
                            state = log.userOutput("Not public", state=state)
                            continue
                        r = s.get(pdf_url, headers = headers, timeout = 5)
                        pdf_real = 'https://ncbi.nlm.nih.gov'+r.html.find('a.int-view', first=True).attrs['href']
                        state = log.userOutput(pdf_real, state=state)
                        r = s.get(pdf_real, stream=True)
                        with open(os.path.join(path, idname + '.pdf'), 'wb') as f:
                            for chunk in r.iter_content(chunk_size = 1024):
                                if chunk:
                                    f.write(chunk)
                    except AttributeError:
                        pass
                        state = log.userOutput(f"{idname} could not be gathered.", state=state)

                except ConnectionError as e:
                    pass
                    state = log.userOutput(f"{idname} could not be gathered.", state=state)
        else:
            state = log.userOutput(f"No paper associated with {identifier} found on PubMed", state=state)


#THIS ONE WORKS BETTER
def annotations(ids, state):
    """
    Downloads annotations for a specified gene product.

    :param ids: The gene product identifier for which the annotations are to be downloaded.
    :type ids: str

    :raises requests.HTTPError: If the HTTP request to download the annotations fails.

    """
    # Auth: Marc Choi
    #       machoi@umich.edu
    try: 
        path = os.path.abspath(os.getcwd()) + '/go_annotations'
        os.makedirs(path, exist_ok = True) 
        state = log.userOutput("Directory '%s' created successfully" % path, state=state)
    except OSError as error: 
        state = log.userOutput("Directory '%s' can not be created" % path, state=state)
        
        
    requestURL = "https://amigo.geneontology.org/amigo/search/annotation?q="+ids
    real_id = ids[3:]
    fetch_annotation(real_id, path)


def fetch_annotation(id, path):
    # URL to fetch data from
    url = 'https://golr-aux.geneontology.io/solr/select?defType=edismax&qt=standard&indent=on&wt=csv&rows=100000&start=0&fl=source%2Cbioentity_internal_id%2Cbioentity_label%2Cqualifier%2Cannotation_class%2Creference%2Cevidence_type%2Cevidence_with%2Caspect%2Cbioentity_name%2Csynonym%2Ctype%2Ctaxon%2Cdate%2Cassigned_by%2Cannotation_extension_class%2Cbioentity_isoform&facet=true&facet.mincount=1&facet.sort=count&json.nl=arrarr&facet.limit=25&hl=true&hl.simple.pre=%3Cem%20class%3D%22hilite%22%3E&hl.snippets=1000&csv.encapsulator=&csv.separator=%09&csv.header=false&csv.mv.separator=%7C&fq=document_category:%22annotation%22&facet.field=aspect&facet.field=taxon_subset_closure_label&facet.field=type&facet.field=evidence_subset_closure_label&facet.field=regulates_closure_label&facet.field=isa_partof_closure_label&facet.field=annotation_class_label&facet.field=qualifier&facet.field=annotation_extension_class_closure_label&facet.field=assigned_by&facet.field=panther_family_label&q=GO%3A'+id+'&qf=annotation_class%5E2&qf=annotation_class_label_searchable%5E1&qf=bioentity%5E2&qf=bioentity_label_searchable%5E1&qf=bioentity_name_searchable%5E1&qf=annotation_extension_class%5E2&qf=annotation_extension_class_label_searchable%5E1&qf=reference_searchable%5E1&qf=panther_family_searchable%5E1&qf=panther_family_label_searchable%5E1&qf=bioentity_isoform%5E1&qf=isa_partof_closure%5E1&qf=isa_partof_closure_label_searchable%5E1'
    print(url)
# Fetch data from the URL
    response = requests.get(url)

# Check if the request was successful
    if response.status_code == 200:
    # Read the data into a pandas DataFrame
        data = StringIO(response.text)
        df = pd.read_csv(data, sep='\t')  # '\t' is the tab separator as specified in the URL
    
    # Define the new column names
        new_column_names = [
            'Source',
            'Bioentity Internal ID',
            'Bioentity Label',
            'Qualifier',
            'Annotation Class',
            'Reference',
            'Evidence Type',
            'Evidence With',
            'Aspect',
            'Bioentity Name',
            'Synonym',
            'Type',
            'Taxon',
            'Date',
            'Assigned By',
            'Annotation Extension Class',
            'Bioentity Isoform'
        ]
    
    # Rename the columns directly
        df.columns = new_column_names
    
    # Save the updated DataFrame to a CSV file
        df.to_csv(path+'/annotations_go_'+id+'.csv', index=False)
        print("Data with updated column names has been saved to "+path+"/annotations_go_"+id+".csv")
    else:
        print(f"Failed to retrieve data. Status code: {response.status_code}")






"""
This module is the access point for having a conversation with the LLM and no other tools.
"""

import pandas as pd
import numpy as np
import chromadb
import time
import subprocess
import os
from langchain.document_loaders import DirectoryLoader
from langchain.document_loaders import UnstructuredPDFLoader
from langchain.vectorstores import Chroma
from langchain.embeddings import HuggingFaceEmbeddings
from langchain.chains import RetrievalQA
from langchain.llms import LlamaCpp
from langchain.callbacks.manager import CallbackManager
from langchain.callbacks.streaming_stdout import StreamingStdOutCallbackHandler
from langchain.document_loaders import PyPDFLoader
# from langchain.text_splitter import CharacterTextSplitter
from langchain.chains.question_answering import load_qa_chain
from langchain.output_parsers import CommaSeparatedListOutputParser
from semantic_router.layer import RouteLayer
from langchain.chains import ConversationChain
from langchain.prompts import PromptTemplate
from langchain_core.prompts.prompt import PromptTemplate
from langchain.docstore.document import Document
from langchain_chroma import Chroma
from langchain_community.document_loaders import PyPDFLoader
from langchain_community.document_loaders import TextLoader
from langchain_community.embeddings.sentence_transformer import (
    SentenceTransformerEmbeddings,
)
# from sklearn.metrics.pairwise import cosine_similarity
# from langchain_text_splitters import CharacterTextSplitter
# from langchain.retrievers.multi_query import MultiQueryRetriever
from langchain_community.callbacks import get_openai_callback
# from langchain_text_splitters import RecursiveCharacterTextSplitter

# from langchain_text_splitters import RecursiveCharacterTextSplitter

# from sentence_transformers import SentenceTransformer, util

import logging

from BRAD.promptTemplates import (
    history_chat_template,
    get_default_context,
)

# Extraction
import re

from BRAD import utils
from BRAD import log


def llm_only(state):
    """
    Executes a conversational interaction using a large language model (LLM) without retrieving external knowledge from a vector database.

    This function sets up and runs a conversational chain that uses a given LLM and a memory module. It constructs the prompt, tracks the LLM's usage (e.g., token count and cost), and logs the details of the interaction. The function also appends the LLM's response to the process log and displays the output to the user.

    Parameters:
        state (dict): A dictionary containing the current execution state, including:
            - 'llm' (object): The large language model to use.
            - 'prompt' (str): The user's input prompt.
            - 'databases' (dict): A dictionary of available databases; 'RAG' represents the vector database.
            - 'memory' (object): The memory module for the conversation.
            - 'config' (dict): Configuration settings, including 'debug' for verbosity.
            - 'process' (dict): A process log tracking the steps taken.

    Returns:
        dict: Updated `state` object with the LLM's response logged and user output appended.
    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: Nov. 17, 2024

    # Get information from Agent.state
    llm = state["llm"]  # get the llm
    prompt = state["prompt"]  # get the user prompt
    vectordb = state["databases"]["RAG"]  # get the vector database
    memory = state["memory"]  # get the memory of the model

    # Get information to invoke the llm
    template = history_chat_template()
    PROMPT = PromptTemplate(input_variables=["history", "input"], template=template)
    conversation = ConversationChain(
        prompt=PROMPT,
        llm=llm,
        verbose=state["config"]["debug"],
        memory=memory,
    )
    prompt = get_default_context() + prompt

    # Invoke LLM tracking its usage
    start_time = time.time()
    with get_openai_callback() as cb:
        response = conversation.predict(input=prompt)

    # Track llm usage
    responseDetails = {
        "content": response,
        "time": time.time() - start_time,
        "call back": {
            "Total Tokens": cb.total_tokens,
            "Prompt Tokens": cb.prompt_tokens,
            "Completion Tokens": cb.completion_tokens,
            "Total Cost (USD)": cb.total_cost,
        },
    }

    # Log the LLM response
    state["process"]["steps"].append(
        log.llmCallLog(
            llm=llm,
            prompt=PROMPT,
            input=prompt,
            output=responseDetails,
            parsedOutput=response,
            apiInfo=responseDetails["call back"],
            purpose="justchat.llm_only",
        )
    )

    # Display output to the user
    state = log.userOutput(response, state=state)

    return state



"""
This module provides utilities for logging and tracking chat processes and interactions 
within the BRAD framework. It includes functions to record the agent state, LLM calls, file loading, 
debug information, and user-facing outputs, and other pieces of information to ensure consistent
and transparent monitoring of the `Agent` and user interactions. These logging methods should be inserted
throughout the code for development, debugging, and  user-facing output. Log files are
automatically saved in the output directory.

Log Format
----------
Logs capture user interactions and actions taken by an `Agent`. These logs are saved in 
the output directory and can be reviewed for debugging or analysis purposes. Within the log
record, entries are numbered according to their order in the chat. For a single entry, the following
items are recorded:

1. time: the time when the `Agent` responds to this query
2. prompt: the input prompt from the user
3. output: the message displayed to the user
4. status: the `Agent` state after responding to the user (See :ref:`state-schema-section`)
5. process: this records the tool module and the set of particular steps the `Agent` takes using the tool
6. planned: a list of any steps that `Agent` plans to take

These items are saved in the following schema:

>>> [
...     0: {
...         'TIME'   : <time stamp>,
...         'PROMPT' : <input from the user or preplanned prompt>,
...         'OUTPUT' : <output displayed to the user>,
...         'STATUS' : {
...                 'LLM' : <primary large language model being used>,
...                 'Databases' : {
...                         'RAG' : <primary data base>,
...                         <other databases>: <more databases can be added>,
...                         ...
...                     },
...                 'config' : {
...                         'debug' : True,
...                         'output-directory' : <path to output directory>,
...                         ...
...                     }
...             },
...         'PROCESS' : {
...                 'MODULE' : <name of module i.e. RAG, DATABASE, CODER, etc.>
...                 'STEPS' [
...                         # information for a particular step involved in executing the module. Some examples
...                         {
...                            'func' : 'rag.retreival',
...                            'num articles retrieved' : 10,
...                            'multiquery' : True,
...                            'compression' : True,
...                         },
...                         {
...                            'LLM' : <language model>,
...                            'prompt template' : <prompt template>,
...                            'model input' : <input to model>,
...                            'model output' : <output of model>
...                         }
...                     ]
...             },
...         'PLANNED' : [
...                 <Next prompt in queue>,
...                 ...
...             ]
...     },
... 1 : {
...         'TIME'   : <time stamp>,
...         'PROMPT' : <second prompt>,
...         ...
...     },
... ...
>>> ]

Log Methods
-----------

The methods in the log module serve three key functions:

1. Saving information to the log file.
2. Ensuring consistency in how data is recorded within the `steps` section of the log (e.g., all LLM calls follow a uniform format).
3. Displaying output, warnings, and error messages to the user.

The following methods are available:
"""

import json
import logging
import time

def logger(chatlog, state, chatname, elapsed_time=None):
    """
    This methods writes the lof of the current chat status, user inputs, outputs, and process details to
    a specified file.

    The function serializes the chat data, including user inputs, outputs, process information, and the current
    `Agent` state. The resulting log is written to a file specified by `chatname`.

    :param chatlog: A dictionary containing the chat log entries. Each entry is a record of a specific chat interaction.
    :type chatlog: dict
    :param state: The current status of the chat, including details like the user's prompt, the system's output, 
                  process information, and other contextual information (e.g., databases, queue).
    :type state: dict
    :param chatname: The filename (including path) where the chat log will be saved. The file will be created or 
                     overwritten if it already exists.
    :type chatname: str
    :param elapsed_time: Optional. The time elapsed since the start of the chat or a specific reference point. 
                         If not provided, it defaults to None.
    :type elapsed_time: float or None, optional

    :raises FileNotFoundError: If the specified chat log file cannot be created or accessed.
    :raises TypeError: If the chat log or status contains non-serializable data that cannot be converted to a string.

    :return: A tuple containing:
             - `chatlog`: The updated chat log with the new entry.
             - `state`: The current state of the system after logging.
    :rtype: tuple

    :note: The log is saved in JSON format with proper indentation to improve readability. Non-serializable data 
           within the chat state (e.g., dataframes) is converted to string format to ensure proper logging.
    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: June 4, 2024
    
    process_serializable = {
            key: value if is_json_serializable(value) else str(value)
            for key, value in state['process'].items()
        }

    chatlog[len(chatlog)] = {
        'time'   : time.ctime(time.time()),
        'elapsed time' : elapsed_time,
        'prompt' : state['prompt'],  # the input to the user
        'output' : state['output'],  # the output to the user
        'continue-module': state.get('continue-module'),
        'process': process_serializable,  # state['process'], # information about the process that ran
        'status' : {                      # information about the chat at that time
            'databases'         : str(state['databases']),
            # 'current table'     : {
            #         'key'       : state['current table']['key'],
            #         'tab'       : state['current table']['tab'].to_json() if state['current table']['tab'] is not None else None,
            #     },
            'current documents' : state['current documents'],
            'queue pointer'     : state['queue pointer'],
            'queue'             : state['queue'].copy()
        },
    }
    with open(chatname, 'w') as fp:
        json.dump(chatlog, fp, indent=4)
    return chatlog, state

def llmCallLog(llm=None, memory=None, prompt=None, input=None, output=None, parsedOutput=None, apiInfo=None, purpose=None):
    """
    Logs the information for each LLM call, capturing relevant details for tracking and debugging. This can
    include both raw and processed outputs, as well as the context of the call.
    
    :param llm: The identifier or name of the LLM used in the call (e.g., GPT-3, GPT-4, etc.). If not provided, 
                it defaults to None.
    :type llm: str, optional
    
    :param memory: A memory object that tracks the conversational history or state associated with the LLM call.
                   This could be a representation of past interactions or context shared between the user and LLM.
    :type memory: object, optional
    
    :param prompt: The template or content used to prompt the LLM. This can be a system prompt or a structured input
                   that guides the LLM in generating a response.
    :type prompt: str, optional
    
    :param input: The full input provided to the LLM, which includes both the prompt and any additional context
                  or parameters. This represents the complete information sent to the model for processing.
    :type input: str, optional
    
    :param output: The complete output returned by the LLM. This may contain large amounts of text or data generated 
                   by the model in response to the input.
    :type output: str, optional
    
    :param parsedOutput: The processed or parsed output that is actually used or displayed. This can be a subset 
                         or reformatted version of the full LLM output, focusing on the most relevant information.
    :type parsedOutput: any, optional

    :param apiInfo: The callback information regarding LLM api utilization and fees

    :type apiInfo: dict, optional
    
    :param purpose: A description of the purpose of the LLM call. This field explains why the LLM was called and
                    what task or goal the call is intended to achieve (e.g., generate response, extract information).
    :type purpose: str, optional
    
    :return: A dictionary containing the logged information for the LLM call, including the LLM name, memory,
             prompt, input, output, parsedOutput, and purpose.
    :rtype: dict
    
    :note: Callback information from OpenAI or NVIDIA APIs can be saved in the output field, which contains the full output
           produced by the LLM. The `parsedOutput` should contain only the information used by BRAD or displayed to the user.
    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: June 19, 2024
    llmLog = {
        'llm'          : str(llm),     # LLM
        'memory'       : str(memory),  # memory object
        'prompt'       : str(prompt),  # prompt template for LLM
        'input'        : str(input),   # full input to LLM
        'output'       : str(output),  # full output from LLM
        'parsedOutput' : parsedOutput, # used output information
        'api-info'     : apiInfo,      # information about LLM api utilization
        'purpose'      : str(purpose)  # why is this llm call needed?
    }
    return llmLog

def loadFileLog(file=None, delimiter=None):
    """
    Logs and returns information about a file that has been loaded into the system. This function captures
    details such as the file name and the delimiter used to parse the file.

    The logged information is returned as a dictionary, making it easy to track and review the file loading 
    process during later stages, such as debugging or auditing the steps involved in data processing.

    :param file: The name or path of the file being loaded. If not provided, it defaults to `None`.
    :type file: str, optional
    :param delimiter: The delimiter used to parse the file (e.g., commas for CSV files). If not specified, 
                      it defaults to `None`.
    :type delimiter: str, optional

    :return: A dictionary containing:
             - `file`: The name or path of the file (as a string).
             - `delimiter`: The delimiter used in the file (as a string).
    :rtype: dict

    :note: This function can be useful for logging metadata about files loaded into the system, especially 
           when working with various file formats that may require different parsing strategies.
    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: June 19, 2024
    loadLog = {
        'file'      : str(file),
        'delimiter' : str(delimiter)
    }
    return loadLog

def debugLog(output, state=None, display=None):
    """
    Records debug information to a log file and optionally displays messages to the user based on the system's 
    debug settings.

    This function is used to capture and report errors or other important debug information. It checks the 
    current debug configuration and either writes the log to a file or displays it to the user, ensuring 
    flexibility in managing debug output.

    :param output: The message or error to be logged. This can be any information that needs to be saved for 
                   debugging or tracking purposes.
    :type output: str
    :param state: The current state of the system, containing configuration details, including whether debug 
                  logging is enabled. This is used to determine whether to log the output.
    :type state: dict, optional
    :param display: If set to `True`, the output will be displayed to the user regardless of the debug configuration. 
                    If `False` or not provided, the function will refer to the debug setting in `state['config']['debug']`.
    :type display: bool, optional

    :raises KeyError: If `state['config']['debug']` is not found when `display` is `None`.
    
    :note: Debug logs are saved using Python's logging library, and the output format includes timestamps and 
           logging levels. If `display` is enabled or debug mode is active, the output will be shown to the user.
           Otherwise, it will be silently logged to the configured log file.
    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: June 19, 2024
    logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
    if display:
        logging.info(output)
    elif state is not None and state['config']['debug']:
        logging.info(output)


def errorLog(errorMessage, info=None, state=None):
    """
    Logs an error message and related information to the log file and updates the process steps in the system's state.

    This function is designed to log errors that occur during the operation of the BRAD framework. It records the 
    error message in the log file with a timestamp, and if provided, additional information (`info`) can be logged 
    as well. The error is also appended to the `state['process']['steps']` list for tracking in the system's state.

    :param errorMessage: The error message to be logged. This is a string that describes the nature of the error.
    :type errorMessage: str
    :param info: Optional additional information related to the error. This can provide context or further details 
                 about the error, such as relevant variables or states at the time of the error.
    :type info: any, optional
    :param state: The current system state, which contains information about the chat process and tracks the 
                  progression of steps. The error and related information will be appended to `state['process']['steps']`.
    :type state: dict, optional

    :raises KeyError: If `state['process']['steps']` is not found, meaning the state has not been properly initialized 
                      for logging steps.

    :note: The error is logged using Python's logging library, which outputs the error message with a timestamp 
           and logging level. Additionally, the error is tracked within the `state` object for review and debugging purposes.
           Make sure that `state['process']['steps']` is correctly initialized before calling this function.
    """
    logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
    logging.info(errorMessage)
    state['process']['steps'].append(
        {
            'func'    : 'log.errorLog',
            'message' : errorMessage,
            'info'    : info
        }
    )
    

def userOutput(output, state=None):
    """
    Standardizes and manages the display of information to the user, while also logging the output.

    This method is responsible for both printing messages directly to the user and saving the output to 
    the `Agent.state` for logging purposes. It ensures that the displayed output is consistently tracked and appended 
    to the system's state for future reference, debugging, or record-keeping. 

    :param output: The message or information to be displayed to the user. This can be any printable string or data 
                   that the user should see.
    :type output: str
    :param state: The current system state, which tracks the chat process and logs the output. If `state['output']` 
                  is not initialized, this method will set it to the given output. If it is already initialized, 
                  the method appends the new output to the existing log.
    :type state: dict, optional

    :return: The updated state object with the latest output appended.
    :rtype: dict

    :raises KeyError: If the state is not initialized properly or lacks an 'output' key, this method attempts to create 
                      or update the 'output' field within the state object.

    :note: This is the only method in the BRAD framework that directly displays output to the user. The logs are saved 
           in the state object to track all output communications, ensuring a record is available for debugging or review. 
           Make sure that the state is correctly initialized before calling this method.
    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: June 20, 2024
          
    print(output)
    if 'output' not in state.keys() or state['output'] is None:
        state['output'] = output
    else:
        state['output'] += output
    
    return state

def is_json_serializable(value):
    """
    Determines whether a given value can be serialized into a valid JSON format.

    This function checks if the provided value can be safely converted to JSON. It is useful for ensuring that data, 
    especially when dealing with logs or state information, can be stored in JSON format without errors. The function 
    handles common data types and catches serialization errors for non-compatible types.

    :param value: The data to be evaluated for JSON serializability. This can be any Python object.
    :type value: Any

    :return: Returns True if the value can be serialized to JSON. Returns False if the value raises a TypeError or 
             OverflowError when attempting serialization.
    :rtype: bool

    :raises TypeError: If the value is of a type that cannot be serialized (e.g., custom objects, functions).
    :raises OverflowError: If the value exceeds the limits of what JSON can represent (e.g., extremely large numbers).

    :note: Lists are treated as JSON serializable by default. For other complex types like dictionaries, sets, or custom 
           objects, serialization will depend on the specific structure and contents.

    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: June 4, 2024

    if isinstance(value, list):
        return True
    
    try:
        json.dumps(value)
        return True
    except (TypeError, OverflowError):
        return False



"""
Code Caller
-----------

This module facilitates the discovery, selection, and execution of Python and MATLAB scripts 
based on user prompts and predefined configuration settings.

Key Features
~~~~~~~~~~~~

1. Python scripts must reside in the directories specified within the configuration settings.

2. Script execution requires the first argument to specify the output directory, where any 
   resulting files will be saved.

3. Each script must include clear and structured documentation, consisting of:

   - A concise one-line summary at the beginning of the docstring (used by the LLM for script selection).
   
   - Comprehensive descriptions detailing the script’s arguments, inputs, purpose, and usage 
     examples (utilized by the LLM for accurate execution).

Available Methods
~~~~~~~~~~~~~~~~~

This module has the following methods:

"""


import os
import time
from langchain import PromptTemplate, LLMChain
from langchain.chains import ConversationChain
from langchain.memory import ConversationBufferMemory
from langchain_community.callbacks import get_openai_callback

from BRAD.pythonCaller import find_py_files, get_py_description, read_python_docstrings, pythonPromptTemplate, extract_python_code, execute_python_code
from BRAD.promptTemplates import scriptSelectorTemplate, pythonPromptTemplateWithFiles, summarize_code_template
from BRAD import log
from BRAD import utils
from BRAD import justchat

# History:
#  2024-10-01: This file was modified to remove support for running MATLAB codes
#  2024-11-27: Final stage of the pipeline added to have the LLM generate a response
#              at the end of the pipeline

def code_caller(state):
    """
    Executes a Python script based on the user's prompt and chat status settings.
    
    This function performs the following steps:
    
        1. Searches the specified directories for available Python scripts.
        
        2. Extracts and analyzes docstrings from each script to identify their purpose.
        
        3. Uses a large language model (LLM) to select the most appropriate script and format the command with the correct inputs.
        
        4. Executes the selected script and updates the chat status accordingly.
    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: June 22, 2024

    log.debugLog("CODER", state=state)
    prompt = state['prompt']                                        # Get the user prompt
    llm = state['llm']                                              # Get the llm
    memory = ConversationBufferMemory(ai_prefix="BRAD")                  # state['memory']

    # Get available matlab and python scripts
    path = state['config']['CODE']['path']
    
    scripts = {}
    for fdir in path:
        scripts[fdir] = {}
        scripts[fdir]['python'] = find_py_files(fdir) # pythonScripts
    #    scripts[fdir]['matlab'] = find_matlab_files(fdir) # matlabScripts
    
    # Get matlab and python docstrings
    scriptPurpose = {}
    for fdir in path:
        # print(fdir)
        for script in scripts[fdir]['python']:
            scriptPurpose[script] = {'description': get_py_description(os.path.join(fdir, script + '.py')), 'type': 'python'}
    script_list = ""
    for script in scriptPurpose.keys():
        script_list += "Script Name: " + script + ", \t Description: " + scriptPurpose[script]['description'] + '\n'

    # Determine which code needs to be executed (first llm call)
    template = scriptSelectorTemplate()
    template = template.format(script_list=script_list)
    PROMPT = PromptTemplate(input_variables=["user_query"], template=template)
    log.debugLog(PROMPT, state=state)
    chain = PROMPT | llm # LCEL chain creation
    log.debugLog("FIRST LLM CALL", state=state)

    # Call LLM
    start_time = time.time()
    with get_openai_callback() as cb:
        res = chain.invoke(prompt)
    responseOriginal = {'original': res.copy()}
    responseOriginal['time'] = time.time() - start_time
    responseOriginal['call back'] = {
        "Total Tokens": cb.total_tokens,
        "Prompt Tokens": cb.prompt_tokens,
        "Completion Tokens": cb.completion_tokens,
        "Total Cost (USD)": cb.total_cost
    }

    log.debugLog(res.content, state=state)
    scriptName   = res.content.strip().split('\n')[0].split(':')[1].strip()
    scriptType   = scriptPurpose[scriptName]['type']
    scriptPath = None
    for fdir in path:
        if scriptName in scripts[fdir][scriptType]:
            scriptPath = fdir
            break
    if scriptPath is None:
        log.debugLog('the scriptPath was not found', state=state)
        log.debugLog(f'scripts={scripts}', state=state)
        log.debugLog(f'scriptName={scriptName}', state=state)
        log.debugLog(f'scriptType={scriptType}', state=state)

    # NOTE: MATLAB is in an experimental stage and not fully integrated yet
    if scriptType == 'MATLAB':
        state, _ = activateMatlabEngine(state) # turn on and add matlab files to path
        scriptName = os.path.join(scriptPath, scriptName)
    else:
        log.debugLog('scriptPath=' + str(scriptPath), state=state)
        scriptName = os.path.join(scriptPath, scriptName)

    scriptSuffix = {'python': '.py', 'MATLAB': '.m'}.get(scriptType)
    scriptName  += scriptSuffix

    state['process']['steps'].append(
        log.llmCallLog(
            llm             = llm,
            prompt          = PROMPT,
            input           = prompt,
            output          = responseOriginal,
            parsedOutput    = {
                'scriptName': scriptName,
                'scriptType': scriptType,
                'scriptPath': scriptPath
            },
            purpose         = 'Select which code to run'
        )
    )

    # Format code to execute: read the doc strings, format function call (second llm call), parse the llm output
    log.debugLog("ALL SCRIPTS FOUND. BUILDING TEMPLATE", state=state)
    
    docstringReader = {'python': read_python_docstrings}.get(scriptType)
    docstrings      = docstringReader(os.path.join(scriptPath, scriptName))
    scriptCallingTemplate = {'python': pythonPromptTemplateWithFiles}.get(scriptType)
    template        = scriptCallingTemplate()
    if scriptType == 'python':
        createdFiles = "\n".join(utils.outputFiles(state)) # A string of previously created files
        filled_template = template.format(
            scriptName=scriptName,
            scriptDocumentation=docstrings,
            output_path=state['output-directory'],
            files=createdFiles
        )
    else:
        filled_template = template.format(
            scriptName=scriptName,
            scriptDocumentation=docstrings,
            output_path=state['output-directory']
        )

    # Create the prompt template
    PROMPT = PromptTemplate(input_variables=["history", "input"], template=filled_template)
    log.debugLog(PROMPT, state=state)
    # this will allow better logging of the response from the query API
    try:
        # LCEL chain creation: prompt | llm
        chain = PROMPT | llm
        
        # Execute the chain with input prompt
        start_time = time.time()
        with get_openai_callback() as cb:
            response = chain.invoke({"history": memory.abuffer(), "input": state['prompt']})

        responseOriginal = response
        responseOriginal['time'] = time.time() - start_time

        try:
            responseOriginal['call back'] = {
                "Total Tokens": cb.total_tokens,
                "Prompt Tokens": cb.prompt_tokens,
                "Completion Tokens": cb.completion_tokens,
                "Total Cost (USD)": cb.total_cost
            }
        except:
           responseOriginal['call back'] = {
                "Total Tokens": None,
                "Prompt Tokens": None,
                "Completion Tokens": None,
                "Total Cost (USD)": None
            }

        response = response.content

    # this catches the initial implementation
    except:
        conversation = ConversationChain(
            prompt  = PROMPT,
            llm     =    llm,
            verbose = state['config']['debug'],
            memory  = memory,
        )
        start_time = time.time()
        with get_openai_callback() as cb:
            response = conversation.predict(input=state['prompt'])

        responseOriginal = response
        try:
            responseOriginal = {
                'content' : response,
                'time' : time.time() - start_time,
                'call back': {
                    "Total Tokens": cb.total_tokens,
                    "Prompt Tokens": cb.prompt_tokens,
                    "Completion Tokens": cb.completion_tokens,
                    "Total Cost (USD)": cb.total_cost
                }
            }
        except:
            responseOriginal = {
                'content' : response,
                'time' : time.time() - start_time,
                'call back': {
                    "Total Tokens": None,
                    "Prompt Tokens": None,
                    "Completion Tokens": None,
                    "Total Cost (USD)": None
                }
            }
    
    responseParser = {'python': extract_python_code}.get(scriptType)
    code2execute = responseParser(response, scriptPath, state, memory=memory)

    state['process']['steps'].append(
        log.llmCallLog(
            llm             = llm,
            prompt          = PROMPT,
            input           = state['prompt'],
            output          = responseOriginal,
            parsedOutput    = {
                'code': code2execute
            },
            purpose         = 'Format function call'
        )
    )

    # Check if it requires previous inputs
    code2execute = utils.add_output_file_path_to_string(code2execute, state)
    
    # Execute code
    executeCode(state, code2execute, scriptType)

    # LLM generated response
    state = summarize_code_execution_results(state)

    return state

def summarize_code_execution_results(state):
    """
    This method takes the output of the code and formulates it into a prompt which can be passed to the LLM.
    The LLM response is generated with the justchat method.
    """
    software_output = state['output']
    
    template = summarize_code_template()
    template = template.format(
        user_query=state['prompt'],
        system_output=state['output']
    )
    state['prompt'] = template

    state = justchat.llm_only(state)
    state['output'] = software_output + '\n\n --- \n\n ' + state['output']

    return state

def executeCode(state, code2execute, scriptType):
    """
    Executes the provided code based on the specified script type.
    
    This function determines the appropriate execution environment (Python or MATLAB) based on the script type and runs the corresponding code.
    
    :param state: A dictionary containing the chat status, including configuration settings and other relevant data.
    :type state: dict
    :param code2execute: The code to be executed.
    :type code2execute: str
    :param scriptType: The type of the script to be executed. Must be either 'python' or 'MATLAB'.
    :type scriptType: str
    """
    # Auth: Joshua Pickard
    #       jpic@umich.edu
    # Date: June 22, 2024
    executor = {'python': execute_python_code}.get(scriptType)
    executor(code2execute, state)

